# -*- coding: UTF-8 -*-
""" Run an experiment given a parameter file. Extracts grammars, does parsing
and evaluation. """
from __future__ import division, print_function
import io
import os
import re
import sys
import time
import gzip
import codecs
import logging
import tempfile
import multiprocessing
from itertools import islice
if sys.version[0] >= '3':
	import pickle
	from itertools import zip_longest  # pylint: disable=E0611
else:
	import cPickle as pickle
	from itertools import izip_longest as zip_longest
from collections import defaultdict, OrderedDict, Counter as multiset
from operator import itemgetter
from subprocess import Popen, PIPE
import numpy as np
from discodop import eval as evalmod
from discodop.tree import Tree
from discodop.treebank import getreader, writetree, treebankfanout
from discodop.treebanktransforms import transform, rrtransform
from discodop.treetransforms import binarize, optimalbinarize, canonicalize, \
		splitdiscnodes, addfanoutmarkers
from discodop.treedraw import DrawTree
from discodop.fragments import getfragments
from discodop.grammar import treebankgrammar, dopreduction, \
		doubledop, grammarinfo, write_lcfrs_grammar, sortgrammar
from discodop._grammar import Grammar
from discodop.lexicon import getunknownwordmodel, getlexmodel, smoothlexicon, \
		simplesmoothlexicon, replaceraretrainwords, getunknownwordfun
from discodop.parser import DEFAULTSTAGE, readgrammars, Parser, DictObj
from discodop.estimates import getestimates, getpcfgestimates

USAGE = """Usage: %s [--rerun] <parameter file>
If a parameter file is given, an experiment is run. See the file sample.prm for
an example parameter file. To repeat an experiment with an existing grammar,
pass the option --rerun.""" % sys.argv[0]

INTERNALPARAMS = None


def initworker(params):
	""" Set global parameter object """
	global INTERNALPARAMS
	INTERNALPARAMS = params


def startexp(
		stages=(DEFAULTSTAGE, ),  # see above
		corpusfmt='export',  # choices: export, discbracket, bracket
		corpusdir='.',
		# filenames may include globbing characters '*' and '?'.
		traincorpus='alpinosample.export', trainencoding='utf-8',
		testcorpus='alpinosample.export', testencoding='utf-8',
		testmaxwords=40,
		trainmaxwords=40,
		trainnumsents=2,
		testnumsents=1,  # number of sentences to parse
		skiptrain=True,  # test set starts after training set
		# (useful when they are in the same file)
		skip=0,  # number of sentences to skip from test corpus
		punct=None,  # choices: None, 'move', 'remove', 'root'
		functions=None,  # choices None, 'add', 'remove', 'replace'
		morphology=None,  # choices: None, 'add', 'replace', 'between'
		transformations=None,  # apply treebank transformations
		# postagging: pass None to use tags from treebank.
		postagging=dict(
			method='unknownword',
			# choices: unknownword (assign during parsing),
			# 		treetagger, stanford (external taggers)
			# choices unknownword: 4, 6, base,
			# for treetagger / stanford: [filename of external tagger model]
			model='4',
			# options for unknown word models:
			unknownthreshold=1,  # use probs of rare words for unknown words
			openclassthreshold=50,  # add unseen tags for known words; 0=disable
			simplelexsmooth=True,  # disable sophisticated smoothing
		),
		relationalrealizational=None,  # do not apply RR-transform
		headrules=None,  # rules for finding heads of constituents
		bintype='binarize',  # choices: binarize, optimal, optimalhead
		factor='right',
		revmarkov=True,
		v=1,
		h=2,
		pospa=False,  # when v > 1, add parent annotation to POS tags?
		markhead=False,  # prepend head to siblings
		leftmostunary=True,  # start binarization with unary node
		rightmostunary=True,  # end binarization with unary node
		tailmarker='',  # with headrules, head is last node and can be marked
		fanout_marks_before_bin=False,
		evalparam='proper.prm',  # EVALB-style parameter file
		quiet=False, reallyquiet=False,  # quiet=no per sentence results
		numproc=1,  # increase to use multiple CPUs; None: use all CPUs.
		resultdir='results',
		rerun=False):
	""" Execute an experiment. """
	assert bintype in ('optimal', 'optimalhead', 'binarize')
	if postagging is not None:
		assert set(postagging).issubset({'method', 'model',
				'unknownthreshold', 'openclassthreshold', 'simplelexsmooth'})
		if postagging['method'] == 'unknownword':
			assert postagging['model'] in ('4', '6', 'base')
			assert postagging['unknownthreshold'] >= 1
			assert postagging['openclassthreshold'] >= 0
		else:
			assert postagging['method'] in ('treetagger', 'stanford')

	if rerun:
		assert os.path.exists(resultdir), (
				'Directory %r does not exist.'
				'--rerun requires a directory '
				'with the grammar(s) of a previous experiment.'
				% resultdir)
	else:
		assert not os.path.exists(resultdir), (
			'Directory %r exists.\n'
			'Use --rerun to parse with existing grammar '
			'and overwrite previous results.' % resultdir)
		os.mkdir(resultdir)

	# Log everything, and send it to stderr, in a format with just the message.
	formatstr = '%(message)s'
	if reallyquiet:
		logging.basicConfig(level=logging.WARNING, format=formatstr)
	elif quiet:
		logging.basicConfig(level=logging.INFO, format=formatstr)
	else:
		logging.basicConfig(level=logging.DEBUG, format=formatstr)

	# log up to INFO to a results log file
	fileobj = logging.FileHandler(filename='%s/output.log' % resultdir)
	fileobj.setLevel(logging.INFO)
	fileobj.setFormatter(logging.Formatter(formatstr))
	logging.getLogger('').addHandler(fileobj)

	corpusreader = getreader(corpusfmt)
	if not rerun:
		corpus = corpusreader(corpusdir, traincorpus, encoding=trainencoding,
				headrules=headrules, headfinal=True, headreverse=False,
				punct=punct, functions=functions, morphology=morphology)
		logging.info('%d sentences in training corpus %s/%s',
				len(corpus.parsed_sents()), corpusdir, traincorpus)
		if isinstance(trainnumsents, float):
			trainnumsents = int(trainnumsents * len(corpus.sents()))
		trees = list(corpus.parsed_sents().values())[:trainnumsents]
		sents = list(corpus.sents().values())[:trainnumsents]
		if transformations:
			trees = [transform(tree, sent, transformations)
					for tree, sent in zip(trees, sents)]
		if relationalrealizational:
			trees = [rrtransform(tree, **relationalrealizational)[0]
					for tree in trees]
		train_tagged_sents = [[(word, tag) for word, (_, tag)
				in zip(sent, sorted(tree.pos()))]
					for tree, sent in zip(trees, sents)]
		blocks = list(corpus.blocks().values())[:trainnumsents]
		assert trees, 'training corpus should be non-empty'
		logging.info('%d training sentences before length restriction',
				len(trees))
		trees, sents, blocks = zip(*[sent for sent in zip(trees, sents, blocks)
			if len(sent[1]) <= trainmaxwords])
		logging.info('%d training sentences after length restriction <= %d',
			len(trees), trainmaxwords)

	testset = corpusreader(corpusdir, testcorpus, encoding=testencoding,
			punct=punct, morphology=morphology, functions=functions)
	gold_sents = testset.tagged_sents()
	test_parsed_sents = testset.parsed_sents()
	if skiptrain:
		skip += trainnumsents
	logging.info('%d sentences in test corpus %s/%s',
			len(testset.parsed_sents()), corpusdir, testcorpus)
	logging.info('%d test sentences before length restriction',
			len(list(gold_sents)[skip:skip + testnumsents]))
	lexmodel = None
	test_tagged_sents = gold_sents
	if postagging and postagging['method'] in ('treetagger', 'stanford'):
		if postagging['method'] == 'treetagger':
			# these two tags are never given by tree-tagger,
			# so collect words whose tag needs to be overriden
			overridetags = ('PTKANT', 'PIDAT')
		elif postagging['method'] == 'stanford':
			overridetags = ('PTKANT', )
		taglex = defaultdict(set)
		for sent in train_tagged_sents:
			for word, tag in sent:
				taglex[word].add(tag)
		overridetagdict = {tag:
			{word for word, tags in taglex.items() if tags == {tag}}
			for tag in overridetags}
		tagmap = {'$(': '$[', 'PAV': 'PROAV'}
		sents_to_tag = OrderedDict((a, b) for a, b
				in islice(gold_sents.items(), skip, skip + testnumsents)
				if len(b) <= testmaxwords),
		test_tagged_sents = externaltagging(postagging['method'],
				postagging['model'], sents_to_tag, overridetagdict, tagmap)
		# give these tags to parser
		usetags = True
	elif postagging and postagging['method'] == 'unknownword' and not rerun:
		postagging['unknownwordfun'] = getunknownwordfun(postagging['model'])
		# get smoothed probalities for lexical productions
		lexresults, msg = getunknownwordmodel(
				train_tagged_sents, postagging['unknownwordfun'],
				postagging['unknownthreshold'],
				postagging['openclassthreshold'])
		logging.info(msg)
		simplelexsmooth = postagging['simplelexsmooth']
		if simplelexsmooth:
			lexmodel = lexresults[2:8]
		else:
			lexmodel, msg = getlexmodel(*lexresults)
			logging.info(msg)
		# NB: knownwords are all words in training set, lexicon is the subset
		# of words that are above the frequency threshold.
		# for training purposes we work with the subset, at test time we exploit
		# the full set of known words from the training set.
		sigs, knownwords, lexicon = lexresults[:3]
		postagging['sigs'], postagging['lexicon'] = sigs, knownwords
		# replace rare train words with signatures
		sents = replaceraretrainwords(train_tagged_sents,
				postagging['unknownwordfun'], lexicon)
		# make sure gold POS tags are not given to parser
		usetags = False
	elif postagging and postagging['method'] == 'unknownword' and rerun:
		usetags = False
	else:
		simplelexsmooth = False
		# give gold POS tags to parser
		usetags = True

	# 0: test sentences as they should be handed to the parser,
	# 1: gold trees for evaluation purposes
	# 2: gold sentence because test sentences may be mangled by unknown word
	#   model
	# 3: blocks from treebank file to reproduce the relevant part of the
	#   original treebank verbatim.
	testset = OrderedDict((a, (test_tagged_sents[a], test_parsed_sents[a],
			gold_sents[a], block)) for a, block
			in islice(testset.blocks().items(), skip, skip + testnumsents)
			if len(test_tagged_sents[a]) <= testmaxwords)
	assert test_tagged_sents, 'test corpus should be non-empty'
	logging.info('%d test sentences after length restriction <= %d',
			len(testset), testmaxwords)

	if rerun:
		trees = []
		sents = []
	toplabels = {tree.label for tree in trees} | {
			test_parsed_sents[n].label for n in testset}
	assert len(toplabels) == 1, 'expected unique ROOT label: %r' % toplabels
	top = toplabels.pop()

	if rerun:
		readgrammars(resultdir, stages, postagging, top)
	else:
		logging.info('read training & test corpus')
		getgrammars(trees, sents, stages, bintype, h, v, factor, tailmarker,
				revmarkov, leftmostunary, rightmostunary, pospa, markhead,
				fanout_marks_before_bin, testmaxwords, resultdir, numproc,
				lexmodel, simplelexsmooth, top)
	evalparam = evalmod.readparam(evalparam)
	evalparam['DEBUG'] = -1
	evalparam['CUTOFF_LEN'] = 40
	deletelabel = evalparam.get('DELETE_LABEL', ())
	deleteword = evalparam.get('DELETE_WORD', ())

	begin = time.clock()
	parser = Parser(stages, transformations=transformations,
			tailmarker=tailmarker, postagging=postagging if postagging
			and postagging['method'] == 'unknownword' else None,
			relationalrealizational=relationalrealizational)
	results = doparsing(parser=parser, testset=testset, resultdir=resultdir,
			usetags=usetags, numproc=numproc, deletelabel=deletelabel,
			deleteword=deleteword, corpusfmt=corpusfmt, morphology=morphology)
	if numproc == 1:
		logging.info('time elapsed during parsing: %gs', time.clock() - begin)
	for result in results[0]:
		nsent = len(result.parsetrees)
		header = (' ' + result.name.upper() + ' ').center(35, '=')
		evalsummary = evalmod.doeval(OrderedDict((a, b.copy(True))
				for a, b in test_parsed_sents.items()), gold_sents,
				result.parsetrees, test_tagged_sents if usetags else gold_sents,
				evalparam)
		coverage = 'coverage: %s = %6.2f' % (
				('%d / %d' % (nsent - result.noparse, nsent)).rjust(
				25 if any(len(a) > evalparam['CUTOFF_LEN']
				for a in gold_sents.values()) else 14),
				100.0 * (nsent - result.noparse) / nsent)
		logging.info('\n'.join(('', header, evalsummary, coverage)))
	return top


def getgrammars(trees, sents, stages, bintype, horzmarkov, vertmarkov, factor,
		tailmarker, revmarkov, leftmostunary, rightmostunary, pospa, markhead,
		fanout_marks_before_bin, testmaxwords, resultdir, numproc,
		lexmodel, simplelexsmooth, top):
	""" Apply binarization and read off the requested grammars. """
	# fixme: this n should correspond to sentence id
	tbfanout, n = treebankfanout(trees)
	logging.info('treebank fan-out before binarization: %d #%d\n%s\n%s',
			tbfanout, n, trees[n], ' '.join(sents[n]))
	# binarization
	begin = time.clock()
	if fanout_marks_before_bin:
		trees = [addfanoutmarkers(t) for t in trees]
	if bintype == 'binarize':
		bintype += ' %s h=%d v=%d %s' % (factor, horzmarkov, vertmarkov,
			'tailmarker' if tailmarker else '')
		for a in trees:
			binarize(a, factor=factor, tailmarker=tailmarker,
					horzmarkov=horzmarkov, vertmarkov=vertmarkov,
					leftmostunary=leftmostunary, rightmostunary=rightmostunary,
					reverse=revmarkov, pospa=pospa,
					headidx=-1 if markhead else None)
	elif bintype == 'optimal':
		trees = [Tree.convert(optimalbinarize(tree))
						for n, tree in enumerate(trees)]
	elif bintype == 'optimalhead':
		trees = [Tree.convert(optimalbinarize(tree, headdriven=True,
				h=horzmarkov, v=vertmarkov)) for n, tree in enumerate(trees)]
	trees = [addfanoutmarkers(t) for t in trees]
	logging.info('binarized %s cpu time elapsed: %gs',
						bintype, time.clock() - begin)
	logging.info('binarized treebank fan-out: %d #%d', *treebankfanout(trees))
	trees = [canonicalize(a).freeze() for a in trees]

	for n, stage in enumerate(stages):
		if stage.split:
			traintrees = [binarize(splitdiscnodes(Tree.convert(a),
					stage.markorigin), childchar=':').freeze() for a in trees]
			logging.info('splitted discontinuous nodes')
		else:
			traintrees = trees
		if stage.mode.startswith('pcfg'):
			assert tbfanout == 1 or stage.split
		backtransform = None
		if stage.dop:
			if stage.usedoubledop:
				# find recurring fragments in treebank,
				# as well as depth-1 'cover' fragments
				fragments = getfragments(traintrees, sents, numproc,
						iterate=stage.iterate, complement=stage.complement)
				xgrammar, backtransform, altweights = doubledop(
						traintrees, fragments)
			else:  # DOP reduction
				xgrammar, altweights = dopreduction(
						traintrees, sents, packedgraph=stage.packedgraph)
			nodes = sum(len(list(a.subtrees())) for a in traintrees)
			if lexmodel and simplelexsmooth:
				newrules = simplesmoothlexicon(lexmodel)
				xgrammar.extend(newrules)
				for weights in altweights.values():
					weights.extend(w for _, w in newrules)

				def wordslast(m):
					""" Sort rules, but lexical rules sorted by word at end """
					(r, yf), _ = xgrammar[m]
					return (yf[0] if r[1] == 'Epsilon' else '',
							'}<' in r[0], r)

				idx = sorted(range(len(xgrammar)), key=wordslast)
				altweights = {name: [weights[m] for m in idx]
						for name, weights in altweights.items()}
				xgrammar = [xgrammar[m] for m in idx]
			elif lexmodel:
				xgrammar = smoothlexicon(xgrammar, lexmodel)
			msg = grammarinfo(xgrammar)
			rules, lexicon = write_lcfrs_grammar(
					xgrammar, bitpar=stage.mode.startswith('pcfg'))
			grammar = Grammar(rules, lexicon, start=top,
					bitpar=stage.mode.startswith('pcfg'))
			for name in altweights:
				grammar.register(u'%s' % name, altweights[name])
			with gzip.open('%s/%s.rules.gz' % (
					resultdir, stage.name), 'wb') as rulesfile:
				rulesfile.write(rules)
			with codecs.getwriter('utf-8')(gzip.open('%s/%s.lex.gz' % (
					resultdir, stage.name), 'wb')) as lexiconfile:
				lexiconfile.write(lexicon)
			logging.info('DOP model based on %d sentences, %d nodes, '
				'%d nonterminals', len(traintrees), nodes, len(grammar.toid))
			logging.info(msg)
			if stage.estimator != 'dop1':
				grammar.switch(u'%s' % stage.estimator)
			_sumsto1 = grammar.testgrammar()
			if stage.usedoubledop:
				# backtransform keys are line numbers to rules file;
				# to see them together do:
				# $ paste <(zcat dop.rules.gz) <(zcat dop.backtransform.gz)
				with gzip.open('%s/%s.backtransform.gz' % (
						resultdir, stage.name), 'w') as out:
					out.writelines('%s\n' % a for a in backtransform)
				if n and stage.prune:
					msg = grammar.getmapping(stages[n - 1].grammar,
						striplabelre=None if stages[n - 1].dop
							else re.compile(b'@.+$'),
						neverblockre=re.compile(b'.+}<'),
						# + stage.neverblockre?
						splitprune=stage.splitprune and stages[n - 1].split,
						markorigin=stages[n - 1].markorigin)
				else:
					# recoverfragments() relies on this mapping to identify
					# binarization nodes
					msg = grammar.getmapping(None,
						striplabelre=None,
						neverblockre=re.compile(b'.+}<'),
						# + stage.neverblockre?
						splitprune=False, markorigin=False)
				logging.info(msg)
			elif n and stage.prune:  # dop reduction
				msg = grammar.getmapping(stages[n - 1].grammar,
					striplabelre=None if stages[n - 1].dop
						and not stages[n - 1].usedoubledop
						else re.compile(b'@[-0-9]+$'),
					neverblockre=re.compile(stage.neverblockre)
						if stage.neverblockre else None,
					splitprune=stage.splitprune and stages[n - 1].split,
					markorigin=stages[n - 1].markorigin)
				if stage.mode == 'dop-rerank':
					grammar.getrulemapping(stages[n - 1].grammar)
				logging.info(msg)
			# write prob models
			np.savez_compressed('%s/%s.probs.npz' % (resultdir, stage.name),
					**{name: mod for name, mod
						in zip(grammar.modelnames, grammar.models)})
		else:  # not stage.dop
			xgrammar = treebankgrammar(traintrees, sents)
			logging.info('induced %s based on %d sentences',
				('PCFG' if tbfanout == 1 or stage.split else 'PLCFRS'),
				len(traintrees))
			if stage.split or os.path.exists('%s/pcdist.txt' % resultdir):
				logging.info(grammarinfo(xgrammar))
			else:
				logging.info(grammarinfo(xgrammar,
						dump='%s/pcdist.txt' % resultdir))
			if lexmodel and simplelexsmooth:
				newrules = simplesmoothlexicon(lexmodel)
				xgrammar.extend(newrules)
				xgrammar = sortgrammar(xgrammar)
			elif lexmodel:
				xgrammar = smoothlexicon(xgrammar, lexmodel)
			rules, lexicon = write_lcfrs_grammar(
					xgrammar, bitpar=stage.mode.startswith('pcfg'))
			grammar = Grammar(rules, lexicon, start=top,
					bitpar=stage.mode.startswith('pcfg'))
			with gzip.open('%s/%s.rules.gz' % (
					resultdir, stage.name), 'wb') as rulesfile:
				rulesfile.write(rules)
			with codecs.getwriter('utf-8')(gzip.open('%s/%s.lex.gz' % (
					resultdir, stage.name), 'wb')) as lexiconfile:
				lexiconfile.write(lexicon)
			_sumsto1 = grammar.testgrammar()
			if n and stage.prune:
				msg = grammar.getmapping(stages[n - 1].grammar,
					striplabelre=None,
					neverblockre=re.compile(stage.neverblockre)
						if stage.neverblockre else None,
					splitprune=stage.splitprune and stages[n - 1].split,
					markorigin=stages[n - 1].markorigin)
				logging.info(msg)

		logging.info('wrote grammar to %s/%s.{rules,lex%s}.gz', resultdir,
				stage.name, ',backtransform' if stage.usedoubledop else '')

		outside = None
		if stage.getestimates == 'SX':
			assert tbfanout == 1 or stage.split, 'SX estimate requires PCFG.'
			logging.info('computing PCFG estimates')
			begin = time.clock()
			outside = getpcfgestimates(grammar, testmaxwords,
					grammar.toid[trees[0].label])
			logging.info('estimates done. cpu time elapsed: %gs',
					time.clock() - begin)
			np.savez('pcfgoutside.npz', outside=outside)
			logging.info('saved PCFG estimates')
		elif stage.useestimates == 'SX':
			assert tbfanout == 1 or stage.split, 'SX estimate requires PCFG.'
			assert stage.mode != 'pcfg', (
					'estimates require agenda-based parser.')
			outside = np.load('pcfgoutside.npz')['outside']
			logging.info('loaded PCFG estimates')
		if stage.getestimates == 'SXlrgaps':
			logging.info('computing PLCFRS estimates')
			begin = time.clock()
			outside = getestimates(grammar, testmaxwords,
					grammar.toid[trees[0].label])
			logging.info('estimates done. cpu time elapsed: %gs',
						time.clock() - begin)
			np.savez('outside.npz', outside=outside)
			logging.info('saved estimates')
		elif stage.useestimates == 'SXlrgaps':
			outside = np.load('outside.npz')['outside']
			logging.info('loaded PLCFRS estimates')
		stage.update(grammar=grammar, backtransform=backtransform,
				outside=outside)


def doparsing(**kwds):
	""" Parse a set of sentences using worker processes. """
	params = DictObj(usetags=True, numproc=None, tailmarker='',
		category=None, deletelabel=(), deleteword=(), corpusfmt='export')
	params.update(kwds)
	goldbrackets = multiset()
	totaltokens = 0
	results = [DictObj(name=stage.name) for stage in params.parser.stages]
	for result in results:
		result.update(elapsedtime=dict.fromkeys(params.testset),
				parsetrees=dict.fromkeys(params.testset), brackets=multiset(),
				tagscorrect=0, exact=0, noparse=0)
	if params.numproc == 1:
		initworker(params)
		dowork = (worker(a) for a in params.testset.items())
	else:
		pool = multiprocessing.Pool(processes=params.numproc,
				initializer=initworker, initargs=(params,))
		dowork = pool.imap_unordered(worker, params.testset.items())
	logging.info('going to parse %d sentences.', len(params.testset))
	# main parse loop over each sentence in test corpus
	for nsent, data in enumerate(dowork, 1):
		sentid, msg, sentresults = data
		sent, goldtree, goldsent, _ = params.testset[sentid]
		logging.debug('%d/%d (%s). [len=%d] %s\n%s', nsent, len(params.testset),
					sentid, len(sent),
					' '.join(a[0] for a in goldsent), msg)
		evaltree = goldtree.copy(True)
		evalmod.transform(evaltree, [w for w, _ in sent], evaltree.pos(),
				dict(evaltree.pos()), params.deletelabel, params.deleteword,
				{}, {})
		goldb = evalmod.bracketings(evaltree, dellabel=params.deletelabel)
		goldbrackets.update((sentid, (label, span)) for label, span
				in goldb.elements())
		totaltokens += sum(1 for _, t in goldsent if t not in params.deletelabel)
		for n, result in enumerate(sentresults):
			results[n].brackets.update((sentid, (label, span)) for label, span
					in result.candb.elements())
			assert (results[n].parsetrees[sentid] is None
					and results[n].elapsedtime[sentid] is None)
			results[n].parsetrees[sentid] = result.parsetree
			results[n].elapsedtime[sentid] = result.elapsedtime
			if result.noparse:
				results[n].noparse += 1
			if result.exact:
				results[n].exact += 1
			results[n].tagscorrect += sum(1 for (_, a), (_, b)
					in zip(goldsent, sorted(result.parsetree.pos()))
					if b not in params.deletelabel and a == b)
			logging.debug(
				'%s cov %5.2f tag %5.2f ex %5.2f lp %5.2f lr %5.2f lf %5.2f%s',
					result.name.ljust(7),
					100 * (1 - results[n].noparse / nsent),
					100 * (results[n].tagscorrect / totaltokens),
					100 * (results[n].exact / nsent),
					100 * evalmod.precision(goldbrackets, results[n].brackets),
					100 * evalmod.recall(goldbrackets, results[n].brackets),
					100 * evalmod.f_measure(goldbrackets, results[n].brackets),
					('' if n + 1 < len(sentresults) else '\n'))
	if params.numproc != 1:
		pool.terminate()
		pool.join()
		del dowork, pool

	writeresults(results, params)
	return results, goldbrackets


def worker(args):
	""" Parse a sentence using a global Parser object,
	and do incremental evaluation.

	:returns: a string with diagnostic information, as well as a list of
		DictObj with the results for each stage. """
	nsent, (sent, goldtree, _, _) = args
	prm = INTERNALPARAMS
	goldevaltree = goldtree.copy(True)
	gpos = goldevaltree.pos()
	gposdict = dict(gpos)
	evalmod.transform(goldevaltree, [w for w, _ in sent],
			gpos, gposdict, prm.deletelabel, prm.deleteword, {}, {})
	goldb = evalmod.bracketings(goldevaltree, dellabel=prm.deletelabel)
	results = []
	msg = ''
	for result in prm.parser.parse([w for w, _ in sent],
			tags=[t for _, t in sent] if prm.usetags else None):
		msg += result.msg
		evaltree = result.parsetree.copy(True)
		evalsent = [w for w, _ in sent]
		cpos = evaltree.pos()
		evalmod.transform(evaltree, evalsent, cpos, gposdict,
				prm.deletelabel, prm.deleteword, {}, {})
		candb = evalmod.bracketings(evaltree, dellabel=prm.deletelabel)
		if goldb and candb:
			prec = evalmod.precision(goldb, candb)
			rec = evalmod.recall(goldb, candb)
			f1score = evalmod.f_measure(goldb, candb)
		else:
			prec = rec = f1score = 0
		if f1score == 1.0:
			exact = True
			msg += '\texact match'
		else:
			exact = False
			msg += '\tLP %5.2f LR %5.2f LF %5.2f\n' % (
					100 * prec, 100 * rec, 100 * f1score)
			if (candb - goldb) or (goldb - candb):
				msg += '\t'
			if candb - goldb:
				msg += 'cand-gold=%s ' % evalmod.strbracketings(candb - goldb)
			if goldb - candb:
				msg += 'gold-cand=%s' % evalmod.strbracketings(goldb - candb)
		msg += '\n'
		result.update(dict(candb=candb, exact=exact))
		results.append(result)
	# visualization of last parse tree; highligh matching POS / bracketings
	highlight = [a for a in evaltree.subtrees()
				if evalmod.bracketing(a) in goldb]
	highlight.extend(a for a in evaltree.subtrees()
				if isinstance(a[0], int) and gpos[a[0]] == cpos[a[0]])
	highlight.extend(range(len(cpos)))
	msg += DrawTree(evaltree, evalsent, abbr=True,
			highlight=highlight,
			).text(
				unicodelines=True, ansi=True)
	return (nsent, msg, results)


def writeresults(results, params):
	""" Write parsing results to files in same format as the original corpus.
	(or export if writer not implemented) """
	ext = {'export': 'export',
			'bracket': 'mrg',
			'discbracket': 'dbr',
			'alpino': 'xml'}
	category = (params.category + '.') if params.category else ''
	if params.corpusfmt == 'alpino':
		corpusfmt = 'export'
		io.open('%s/%sgold.%s' % (params.resultdir, category, ext[corpusfmt]),
				'w', encoding='utf-8').writelines(
				writetree(goldtree, [w for w, _ in goldsent], n, corpusfmt,
					morph=params.morphology)
			for n, (_, goldtree, goldsent, _) in params.testset.items())
	else:
		corpusfmt = params.corpusfmt
		io.open('%s/%sgold.%s' % (params.resultdir, category, ext[corpusfmt]),
				'w', encoding='utf-8').writelines(
				a for _, _, _, a in params.testset.values())
	for res in results:
		io.open('%s/%s%s.%s' % (params.resultdir, category, res.name,
				ext[corpusfmt]), 'w', encoding='utf-8').writelines(
				writetree(res.parsetrees[n], [w for w, _ in goldsent], n,
				corpusfmt, morph=params.morphology) for n, (_, _, goldsent, _) in
				params.testset.items())
	with open('%s/parsetimes.txt' % params.resultdir, 'w') as out:
		out.write('#id\tlen\t%s\n' % '\t'.join(res.name for res in results))
		out.writelines('%s\t%d\t%s\n' % (n, len(params.testset[n][2]),
				'\t'.join(str(res.elapsedtime[n]) for res in results))
				for n in params.testset)
	logging.info('wrote results to %s/%s%s.%s', params.resultdir, category,
			(('{%s}' % ','.join(res.name for res in results))
			if len(results) > 1 else results[0].name),
			ext[corpusfmt])


def oldeval(results, goldbrackets):
	""" Simple evaluation. """
	nsent = len(results[0].parsetrees)
	if nsent == 0:
		return
	for result in results:
		logging.info('%s lp %5.2f lr %5.2f lf %5.2f\n'
			'coverage %d / %d = %5.2f %%  exact match %d / %d = %5.2f %%\n',
				result.name,
				100 * evalmod.precision(goldbrackets, result.brackets),
				100 * evalmod.recall(goldbrackets, result.brackets),
				100 * evalmod.f_measure(goldbrackets, result.brackets),
				nsent - result.noparse, nsent,
				100 * (nsent - result.noparse) / nsent,
				result.exact, nsent, 100 * result.exact / nsent)


def readtepacoc():
	""" Read the tepacoc test set. """
	tepacocids = set()
	tepacocsents = defaultdict(list)
	cat = 'undefined'
	tepacoc = io.open('../tepacoc.txt', encoding='utf8')
	for line in tepacoc.read().splitlines():
		fields = line.split('\t')  # = [id, '', sent]
		if line.strip() and len(fields) == 3:
			if fields[0].strip():
				# subtract one because our ids are zero-based, tepacoc 1-based
				sentid = int(fields[0]) - 1
				tepacocids.add(sentid)
				tepacocsents[cat].append((sentid, fields[2].split()))
			else:  # new category
				cat = fields[2]
				if cat.startswith('CUC'):
					cat = 'CUC'
		elif fields[0] == 'TuBa':
			break
	return tepacocids, tepacocsents


def parsetepacoc(
		stages=(dict(mode='pcfg', split=True, markorigin=True),
				dict(mode='plcfrs', prune=True, k=10000, splitprune=True),
				dict(mode='plcfrs', prune=True, k=5000, dop=True,
					usedoubledop=True, estimator='dop1', objective='mpp',
					sample=False, kbest=True)),
		trainmaxwords=999, trainnumsents=25005, testmaxwords=999,
		bintype='binarize', h=1, v=1, factor='right', tailmarker='',
		markhead=False, revmarkov=False, pospa=False,
		leftmostunary=True, rightmostunary=True,
		fanout_marks_before_bin=False, transformations=None,
		usetagger='stanford', resultdir='tepacoc', numproc=1):
	""" Parse the tepacoc test set. """
	for stage in stages:
		for key in stage:
			assert key in DEFAULTSTAGE, 'unrecognized option: %r' % key
	stages = [DictObj({k: stage.get(k, v) for k, v in DEFAULTSTAGE.items()})
			for stage in stages]
	os.mkdir(resultdir)
	# Log everything, and send it to stderr, in a format with just the message.
	formatstr = '%(message)s'
	logging.basicConfig(level=logging.DEBUG, format=formatstr)
	# log up to INFO to a results log file
	fileobj = logging.FileHandler(filename='%s/output.log' % resultdir)
	fileobj.setLevel(logging.INFO)
	fileobj.setFormatter(logging.Formatter(formatstr))
	logging.getLogger('').addHandler(fileobj)
	tepacocids, tepacocsents = readtepacoc()
	try:
		(corpus_sents, corpus_taggedsents,
				corpus_trees, corpus_blocks) = pickle.load(
					gzip.open('tiger.pickle.gz', 'rb'))
	except IOError:  # file not found
		corpus = getreader('export')('../tiger/corpus',
				'tiger_release_aug07.export',
				headrules='negra.headrules' if bintype == 'binarize' else None,
				headfinal=True, headreverse=False, punct='move',
				encoding='iso-8859-1')
		corpus_sents = list(corpus.sents().values())
		corpus_taggedsents = list(corpus.tagged_sents().values())
		corpus_trees = list(corpus.parsed_sents().values())
		if transformations:
			corpus_trees = [transform(tree, sent, transformations)
					for tree, sent in zip(corpus_trees, corpus_sents)]
		corpus_blocks = list(corpus.blocks().values())
		pickle.dump((corpus_sents, corpus_taggedsents, corpus_trees,
			corpus_blocks), gzip.open('tiger.pickle.gz', 'wb'), protocol=-1)

	# test sets (one for each category)
	testsets = {}
	allsents = []
	for cat, catsents in tepacocsents.items():
		testset = sents, trees, goldsents, blocks = [], [], [], []
		for n, sent in catsents:
			if sent != corpus_sents[n]:
				logging.error(
						'mismatch. sent %d:\n%r\n%r\n'
						'not in corpus %r\nnot in tepacoc %r',
						n + 1, sent, corpus_sents[n],
						[a for a, b in zip_longest(sent, corpus_sents[n])
							if a and a != b],
						[b for a, b in zip_longest(sent, corpus_sents[n])
							if b and a != b])
			elif len(corpus_sents[n]) <= testmaxwords:
				sents.append(corpus_taggedsents[n])
				trees.append(corpus_trees[n])
				goldsents.append(corpus_taggedsents[n])
				blocks.append(corpus_blocks[n])
		allsents.extend(sents)
		logging.info('category: %s, %d of %d sentences',
				cat, len(testset[0]), len(catsents))
		testsets[cat] = testset
	testsets['baseline'] = zip(*[sent for n, sent in
				enumerate(zip(corpus_taggedsents, corpus_trees,
						corpus_taggedsents, corpus_blocks))
				if len(sent[1]) <= trainmaxwords
				and n not in tepacocids][trainnumsents:trainnumsents + 2000])
	allsents.extend(testsets['baseline'][0])

	if usetagger:
		overridetags = ('PTKANT', 'VAIMP')
		taglex = defaultdict(set)
		for sent in corpus_taggedsents[:trainnumsents]:
			for word, tag in sent:
				taglex[word].add(tag)
		overridetagdict = {tag:
			{word for word, tags in taglex.items()
			if tags == {tag}} for tag in overridetags}
		tagmap = {'$(': '$[', 'PAV': 'PROAV', 'PIDAT': 'PIAT'}
		# the sentences in the list allsents are modified in-place so that
		# the relevant copy in testsets[cat][0] is updated as well.
		externaltagging(usetagger, '', allsents, overridetagdict, tagmap)

	# training set
	trees, sents, blocks = zip(*[sent for n, sent in
				enumerate(zip(corpus_trees, corpus_sents,
							corpus_blocks)) if len(sent[1]) <= trainmaxwords
							and n not in tepacocids][:trainnumsents])
	getgrammars(trees, sents, stages, bintype, h, v, factor, tailmarker,
			revmarkov, leftmostunary, rightmostunary, pospa, markhead,
			fanout_marks_before_bin, testmaxwords, resultdir,
			numproc, None, False, trees[0].label)
	del corpus_sents, corpus_taggedsents, corpus_trees, corpus_blocks
	results = {}
	cnt = 0
	parser = Parser(stages, tailmarker=tailmarker,
			transformations=transformations)
	for cat, testset in sorted(testsets.items()):
		if cat == 'baseline':
			continue
		logging.info('category: %s', cat)
		begin = time.clock()
		results[cat] = doparsing(parser=parser, testset=testset,
				resultdir=resultdir, usetags=True, numproc=numproc,
				category=cat)
		cnt += len(testset[0])
		if numproc == 1:
			logging.info('time elapsed during parsing: %g',
					time.clock() - begin)
		#else:  # wall clock time here
	goldbrackets = multiset()
	totresults = [DictObj(name=stage.name) for stage in stages]
	for result in totresults:
		result.elapsedtime = [None] * cnt
		result.parsetrees = [None] * cnt
		result.brackets = multiset()
		result.exact = result.noparse = 0
	goldblocks = []
	goldsents = []
	for cat, res in results.items():
		logging.info('category: %s', cat)
		goldbrackets |= res[2]
		goldblocks.extend(res[3])
		goldsents.extend(res[4])
		for result, totresult in zip(res[0], totresults):
			totresult.exact += result.exact
			totresult.noparse += result.noparse
			totresult.brackets |= result.brackets
			totresult.elapsedtime.extend(result.elapsedtime)
		oldeval(*res)
	logging.info('TOTAL')
	oldeval(totresults, goldbrackets)
	# write TOTAL results file with all tepacoc sentences (not the baseline)
	for stage in stages:
		open('TOTAL.%s.export' % stage.name, 'w').writelines(
				open('%s.%s.export' % (cat, stage.name)).read()
				for cat in list(results) + ['gold'])
	# do baseline separately because it shouldn't count towards the total score
	cat = 'baseline'
	logging.info('category: %s', cat)
	oldeval(*doparsing(parser=parser, testset=testsets[cat],
			resultdir=resultdir, usetags=True, numproc=numproc, category=cat))


def externaltagging(usetagger, model, sents, overridetag, tagmap):
	""" Use an external tool to tag a list of tagged sentences. """
	logging.info("Start tagging.")
	goldtags = [t for sent in sents.values() for _, t in sent]
	if usetagger == "treetagger":  # Tree-tagger
		installation = """tree tagger not found. commands to install:
mkdir tree-tagger && cd tree-tagger
wget ftp://ftp.ims.uni-stuttgart.de/pub/corpora/tree-tagger-linux-3.2.tar.gz
tar -xzf tree-tagger-linux-3.2.tar.gz
wget ftp://ftp.ims.uni-stuttgart.de/pub/corpora/tagger-scripts.tar.gz
tar -xzf ftp://ftp.ims.uni-stuttgart.de/pub/corpora/tagger-scripts.tar.gz
mkdir lib && cd lib && wget \
ftp://ftp.ims.uni-stuttgart.de/pub/corpora/german-par-linux-3.2-utf8.bin.gz
gunzip german-par-linux-3.2-utf8.bin.gz"""
		assert os.path.exists("tree-tagger/bin/tree-tagger"), installation
		infile, inname = tempfile.mkstemp(text=True)
		with os.fdopen(infile, 'w') as infile:
			for tagsent in sents.values():
				sent = map(itemgetter(0), tagsent)
				infile.write("\n".join(w.encode('utf-8')
					for n, w in enumerate(sent)) + "\n<S>\n")
		if not model:
			model = "tree-tagger/lib/german-par-linux-3.2-utf8.bin"
			filtertags = "| tree-tagger/cmd/filter-german-tags"
		else:
			filtertags = ''
		tagger = Popen("tree-tagger/bin/tree-tagger -token -sgml"
				" %s %s %s" % (model, inname, filtertags),
				stdout=PIPE, shell=True)
		tagout = tagger.stdout.read(  # pylint: disable=E1101
				).decode('utf-8').split("<S>")[:-1]
		os.unlink(inname)
		taggedsents = OrderedDict((n, [tagmangle(a, None, overridetag, tagmap)
					for a in tags.splitlines() if a.strip()])
					for n, tags in zip(sents, tagout))
	elif usetagger == "stanford":  # Stanford Tagger
		install = """Stanford tagger not found. Commands to install:
wget http://nlp.stanford.edu/software/stanford-postagger-full-2012-07-09.tgz
tar -xzf stanford-postagger-full-2012-07-09.tgz"""
		assert os.path.exists("stanford-postagger-full-2012-07-09"), install
		infile, inname = tempfile.mkstemp(text=True)
		with os.fdopen(infile, 'w') as infile:
			for tagsent in sents.values():
				sent = map(itemgetter(0), tagsent)
				infile.write(' '.join(w.encode('utf-8')
					for n, w in enumerate(sent)) + "\n")
		if not model:
			model = "models/german-hgc.tagger"
		tagger = Popen(args=(
				"/usr/bin/java -mx2G -classpath stanford-postagger.jar"
				" edu.stanford.nlp.tagger.maxent.MaxentTagger"
				" -tokenize false -encoding utf-8"
				" -model %s -textFile %s" % (model, inname)).split(),
				cwd="stanford-postagger-full-2012-07-09",
				shell=False, stdout=PIPE)
		tagout = tagger.stdout.read(  # pylint: disable=E1101
				).decode('utf-8').splitlines()
		os.unlink(inname)
		taggedsents = OrderedDict((n, [tagmangle(a, "_", overridetag, tagmap)
			for a in tags.split()]) for n, tags in zip(sents, tagout))
	assert len(taggedsents) == len(sents), (
			"mismatch in number of sentences after tagging.")
	for n, tags in taggedsents.items():
		assert len(sents[n]) == len(tags), (
				"mismatch in number of tokens after tagging.\n"
				"before: %r\nafter: %r" % (sents[n], tags))
	newtags = [t for sent in taggedsents.values() for _, t in sent]
	logging.info("Tag accuracy: %5.2f\ngold - cand: %r\ncand - gold %r",
		(100 * evalmod.accuracy(goldtags, newtags)),
		set(goldtags) - set(newtags), set(newtags) - set(goldtags))
	return taggedsents

SENTEND = "(\"'!?..."  # ";/-"


def tagmangle(a, splitchar, overridetag, tagmap):
	""" Function to filter tags after they are produced by the tagger. """
	word, tag = a.rsplit(splitchar, 1)
	for newtag in overridetag:
		if word in overridetag[newtag]:
			tag = newtag
	return word, tagmap.get(tag, tag)


def readparam(filename):
	""" Parse a parameter file:
		a list of of attribute=value pairs treated as a dict(). """
	paramstr = open(filename).read()
	params = eval("dict(%s)" % paramstr)
	for stage in params['stages']:
		for key in stage:
			assert key in DEFAULTSTAGE, "unrecognized option: %r" % key
	params['stages'] = [DictObj({k: stage.get(k, v)
			for k, v in DEFAULTSTAGE.items()})
				for stage in params['stages']]
	for n, stage in enumerate(params['stages']):
		assert stage.mode in (
				'plcfrs', 'pcfg', 'pcfg-posterior', 'pcfg-symbolic',
				'pcfg-bitpar', 'dop-rerank')
		assert n > 0 or not stage.prune, (
				"need previous stage to prune, but this stage is first.")
		if stage.mode == 'dop-rerank':
			assert stage.prune and not stage.splitprune and stage.k > 1
			assert (stage.dop and not stage.usedoubledop
					and stage.objective == 'mpp')
		if stage.dop:
			assert stage.estimator in ('dop1', 'ewe', 'bon')
			assert stage.objective in ('mpp', 'mpd', 'shortest',
					"sl-dop", "sl-dop-simple")
	return params


def main(argv=None):
	""" Parse command line arguments. """
	try:
		# report backtrace on segfaults &c. pip install faulthandler
		import faulthandler
		faulthandler.enable()
	except ImportError:
		pass
	if argv is None:
		argv = sys.argv
	if len(argv) == 1:
		print(USAGE)
	elif '--tepacoc' in argv:
		parsetepacoc()
	else:
		params = readparam(argv[1])
		resultdir = argv[1].rsplit('.', 1)[0]
		top = startexp(resultdir=resultdir, rerun='--rerun' in argv, **params)
		if 'rerun' not in argv:  # copy parameter file to result dir
			open("%s/params.prm" % resultdir, "w").write(
					"top='%s',\n%s" % (top, open(argv[1]).read()))



# -*- coding: utf-8 -*-
""" Zhang-Shasha Tree Edit Distance Implementation is licensed under a BSD
style license

Copyright (c) 2012, Tim Henderson and Stephen Johnson
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation
      and/or other materials provided with the distribution.
    * Neither the name of this software nor the names of its contributors may
      be used to endorse or promote products derived from this software without
      specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. """
# andreasvc:
# - modified to work with NLTK Tree objects.
# - added implementation as in Billie (2005), which records edit script.

from __future__ import division, print_function
from collections import deque
from discodop.tree import Tree


class Terminal:
	""" Auxiliary class to be able to add indices to terminal nodes of NLTK
	trees. """
	def __init__(self, node):
		self.prod = self.label = node

	def __repr__(self):
		return repr(self.label)

	def __hash__(self):
		return hash(self.label)

	def __iter__(self):
		return iter(())

	def __len__(self):
		return 0

	def __index__(self):
		return self.label

	def __getitem__(self, val):
		if isinstance(val, slice):
			return ()
		else:
			raise IndexError("A terminal has zero children.")


def prepare(tree, includeterms=False):
	""" Copy a tree object; sort children to have canonical order; merge
	preterminals and terminals in single nodes (unless ``includeterms=True``).
	"""
	tree = tree.copy(True)
	# canonical order of children
	for a in tree.subtrees(lambda n: isinstance(n[0], Tree)):
		a.sort(key=lambda n: min(n.leaves()))
	if includeterms:
		for a in tree.treepositions('leaves'):
			tree[a] = Terminal(tree[a])
	else:
		for a in tree.treepositions('leaves'):
			tree[a[:-1]].label += "-%d" % tree[a]
			del tree[a]
	return tree


# begin Zhang-Shasha Tree Edit Distance Implementation.
class AnnotatedTree:
	""" Wrap a tree to add some extra information. """
	def __init__(self, root):
		self.root = root
		self.nodes = []  # a pre-order enumeration of the nodes in the tree
		self.leftmostdescendents = []  # left most descendents
		self.keyroots = None
		# k and k' are nodes specified in the pre-order enumeration.
		# keyroots = {k | there exists no k'>k such that lmd(k) == lmd(k')}
		# see paper for more on keyroots

		stack = [(root, deque())]
		pstack = []
		j = 0
		while stack:
			n, anc = stack.pop()
			n.id = j
			for c in n:
				a = deque(anc)
				a.appendleft(n.id)
				stack.append((c, a))
			pstack.append((n, anc))
			j += 1
		leftmostdescendents = dict()
		keyroots = dict()
		i = 0
		while pstack:
			n, anc = pstack.pop()
			self.nodes.append(n)
			if len(n) == 0:
				lmd = i
				for a in anc:
					if a in leftmostdescendents:
						break
					else:
						leftmostdescendents[a] = i
			else:
				lmd = leftmostdescendents[n.id]
			self.leftmostdescendents.append(lmd)
			keyroots[lmd] = i
			i += 1
		self.keyroots = sorted(keyroots.values())


def strdist(a, b):
	""" Default categorical distance function. """
	return 0 if a == b else 1


def treedist(tree1, tree2, debug=False):
	""" Zhang-Shasha tree edit distance. """
	tree1 = AnnotatedTree(prepare(tree1))
	tree2 = AnnotatedTree(prepare(tree2))
	tree1lmd = tree1.leftmostdescendents
	tree2lmd = tree2.leftmostdescendents
	tree1nodes = tree1.nodes
	tree2nodes = tree2.nodes
	import numpy
	treedists = numpy.zeros((len(tree1.nodes), len(tree2.nodes)), int)
	for i in tree1.keyroots:
		for j in tree2.keyroots:
			m = i - tree1lmd[i] + 2
			n = j - tree2lmd[j] + 2
			table = numpy.zeros((m, n), int)
			ioff = tree1lmd[i] - 1
			joff = tree2lmd[j] - 1

			for x in range(1, m):  # δ(l(i1)..i, θ) = δ(l(1i)..1-1, θ) + γ(v → λ)
				table[x, 0] = table[x - 1, 0] + 1
			for y in range(1, n):  # δ(θ, l(j1)..j) = δ(θ, l(j1)..j-1) + γ(λ → w)
				table[0, y] = table[0, y - 1] + 1

			for x in range(1, m):
				for y in range(1, n):
					# only need to check if x is an ancestor of i
					# and y is an ancestor of j
					if (tree1lmd[i] == tree1lmd[x + ioff]
							and tree2lmd[j] == tree2lmd[y + joff]):
						#                 +-
						#                 | δ(l(i1)..i-1, l(j1)..j) + γ(v → λ)
						#δ(F1 , F2) = min-+ δ(l(i1)..i , l(j1)..j-1) + γ(λ → w)
						#                 | δ(l(i1)..i-1, l(j1)..j-1) + γ(v → w)
						#                 +-
						labeldist = strdist(tree1nodes[x + ioff].label,
								tree2nodes[y + joff].label)
						table[x, y] = min(table[x - 1, y] + 1,
								table[x, y - 1] + 1,
								table[x - 1, y - 1] + labeldist)
						treedists[x + ioff, y + joff] = table[x, y]
					else:
						#                 +-
						#                 | δ(l(i1)..i-1, l(j1)..j) + γ(v → λ)
						#δ(F1 , F2) = min-+ δ(l(i1)..i , l(j1)..j-1) + γ(λ → w)
						#                 | δ(l(i1)..l(i)-1, l(j1)..l(j)-1)
						#                 |                   + treedist(i1,j1)
						#                 +-
						a = tree1lmd[x + ioff] - 1 - ioff
						b = tree2lmd[y + joff] - 1 - joff
						table[x, y] = min(table[x - 1, y] + 1,
								table[x, y - 1] + 1,
								table[a, b] + treedists[x + ioff, y + joff])
		if debug:
			if isinstance(tree1nodes[i], Tree):
				astr = tree1nodes[i].label  # pprint()
			else:
				astr = str(tree1nodes[i])
			j = treedists[i].argmin()
			if isinstance(tree2nodes[j], Tree):
				bstr = tree2nodes[j].label  # pprint()
			else:
				bstr = str(tree2nodes[j])
			if treedists[i, j]:
				print("%s[%d] %s[%d] %d" % (astr, i, bstr, j, treedists[i, j]))
	return treedists[len(tree1.nodes) - 1, len(tree2.nodes) - 1]
# end Zhang-Shasha Tree Edit Distance Implementation.


def newtreedist(tree1, tree2, debug=False):
	""" implementation as in Billie (2005), based on rparse code.
	slower but records edit script. should be converted to use a set of
	matrices as dynamic programming tables. """
	tree1 = prepare(tree1).freeze()
	tree2 = prepare(tree2).freeze()
	for n, a in enumerate(tree1.subtrees()):
		a.idx = n
	for n, a in enumerate(tree2.subtrees()):
		a.idx = n
	result = geteditstats((tree1,), (tree2,))
	geteditstats.mem.clear()
	if debug:
		print(result)
	return result.distance


class EditStats(object):
	""" Collect edit operations on a tree. """
	__slots__ = ('distance', 'matched', 'editscript')

	def __init__(self, distance=0, matched=0, editscript=None):
		if editscript is None:
			editscript = ()
		self.distance = distance
		self.matched = matched
		self.editscript = editscript

	def __add__(self, other):
		return EditStats(self.distance + other.distance,
			self.matched + other.matched,
			self.editscript + other.editscript)

	def __lt__(self, other):
		return self.distance < other.distance

	def __repr__(self):
		return "%s(distance=%d, matched=%d, [\n\t%s])" % (
				self.__class__.__name__, self.distance, self.matched,
				",\n\t".join("%s(%s, %s)" % (a[0],
					"%s[%d]" % (a[1].label, a[1].idx)
						if isinstance(a[1], Tree) else a[1],
					"%s[%d]" % (a[2].label, a[2].idx)
						if isinstance(a[2], Tree) else a[2])
					for a in self.editscript))


def geteditstats(forest1, forest2):
	""" Recursively get edit distance. """
	try:
		return geteditstats.mem[forest1, forest2]
	except KeyError:
		pass
	flatforest1 = forest1 if forest1 == () else (
			forest1[:-1] + tuple(forest1[-1][:]))
	flatforest2 = forest2 if forest2 == () else (
			forest2[:-1] + tuple(forest2[-1][:]))
	if forest2 == ():
		if forest1 == ():
			result = EditStats(0, 0, ())
		else:
			tmp = geteditstats(flatforest1, ())
			result = EditStats(tmp.distance + 1, tmp.matched,
				(('D', forest1[-1], None), ) + tmp.editscript)
	elif forest1 == ():
		tmp = geteditstats((), flatforest2)
		result = EditStats(tmp.distance + 1, tmp.matched,
			(('I', None, forest2[-1]), ) + tmp.editscript)
	else:
		node1 = forest1[-1]
		node2 = forest2[-1]
		tmp = geteditstats(flatforest1, forest2)
		deletestats = EditStats(tmp.distance + 1, tmp.matched,
				(('D', node1, None),) + tmp.editscript)
		tmp = geteditstats(forest1, flatforest2)
		insertstats = EditStats(tmp.distance + 1, tmp.matched,
				(('I', None, node2), ) + tmp.editscript)
		matchorswapstats = (geteditstats(tuple(node1[:]), tuple(node2[:]))
				+ geteditstats(forest1[:-1], forest2[:-1]))
		if node1.label == node2.label:
			matchorswapstats = EditStats(matchorswapstats.distance,
				matchorswapstats.matched + 1, matchorswapstats.editscript)
		else:
			matchorswapstats = EditStats(matchorswapstats.distance + 1,
				matchorswapstats.matched,
				(('S', node1, node2), ) + matchorswapstats.editscript)
		result = min(deletestats, insertstats, matchorswapstats)
	geteditstats.mem[forest1, forest2] = result
	return result
geteditstats.mem = {}


def main():
	""" Simple demonstration. """
	a = Tree.parse("(f (d (a 0) (c (b 1))) (e 2))", parse_leaf=int)
	b = Tree.parse("(f (c (d (a 0) (b 1)) (e 2)))", parse_leaf=int)
	result1 = treedist(a, b, debug=True)
	assert result1 == 2
	print('%s\n%s\ndistance: %d' % (a, b, result1))
	result2 = newtreedist(a, b, debug=True)
	assert result2 == 2
	print('%s\n%s\ndistance: %d' % (a, b, result2))
	a = Tree.parse("(f (d (x (a 0)) (b 1) (c 2)) (z 3))", parse_leaf=int)
	b = Tree.parse("(f (c (d (a 0) (x (b 1)) (c 2)) (z 3)))", parse_leaf=int)
	result1 = treedist(a, b, debug=True)
	assert result1 == 3
	print('%s\n%s\ndistance: %d' % (a, b, result1))
	result2 = newtreedist(a, b, debug=True)
	assert result2 == 3
	print('%s\n%s\ndistance: %d' % (a, b, result2))

if __name__ == '__main__':
	main()



""" Generic setup.py for Cython code. """
from distutils.core import setup
from Cython.Build import cythonize
from Cython.Distutils import build_ext
from Cython.Compiler import Options
import numpy

metadata = dict(name='disco-dop',
		version='0.3pre1',
		description='Discontinuous Data-Oriented Parsing',
		long_description=open('README.rst').read(),
		author='Andreas van Cranenburgh',
		author_email='A.W.vanCranenburgh@uva.nl',
		url='https://github.com/andreasvc/disco-dop/',
		classifiers=[
				'Development Status :: 4 - Beta',
				'Environment :: Console',
				'Environment :: Web Environment',
				'Intended Audience :: Science/Research',
				'License :: OSI Approved :: GNU General Public License (GPL)',
				'Operating System :: POSIX',
				'Programming Language :: Python',
				'Programming Language :: Python :: 3',
				'Programming Language :: Cython',
				'Topic :: Text Processing :: Linguistic'],
		requires=[
				'cython (>=0.18)',
				'numpy',
		],
		packages=['discodop'],
		scripts=['bin/discodop'],
)

# some of these directives increase performance,
# but at the cost of failing in mysterious ways.
directives = dict(
		profile=False,
		cdivision=True,
		nonecheck=False,
		wraparound=False,
		boundscheck=False,
		embedsignature=True,
)

Options.fast_fail = True
#Options.extra_compile_args = ["-O3"]
#Options.extra_link_args = ["-O3"]  #["-g"],
if __name__ == '__main__':
	setup(
			include_dirs = [numpy.get_include()],
			cmdclass = dict(build_ext=build_ext),
			ext_modules = cythonize(
					'discodop/*.pyx',
					#nthreads=4,
					annotate=True,
					compiler_directives=directives,
					#language_level=3, # FIXME make this work ...
			),
			#test_suite = 'tests'
			**metadata)



""" Priority Queue based on binary heap which implements decrease-key and
remove by marking entries as invalid. Provides dictionary-like interface. Based
on notes in the documentation for heapq, see:
http://docs.python.org/library/heapq.html

There is a version specialized to be used as agenda with edges. """

from __future__ import print_function
from operator import itemgetter

DEF INVALID = 0
DEF HEAP_ARITY = 4
# 2 for binary heap, 4 for quadtree heap


def getkey(Entry entry):
	""" Get key of entry for non-Cython code. """
	return entry.key


cdef inline bint cmpfun(Entry a, Entry b):
	""" Generic comparison function for Entry objects. """
	return (a.value < b.value or (a.value == b.value and a.count < b.count))


cdef inline bint edgecmpfun(Entry a, Entry b):
	""" Comparison function for Entry objects containing Edges. This is
	_significantly_ faster than going through __richcmp__ of objects. """
	return ((<LCFRSEdge>a.value).score < (<LCFRSEdge>b.value).score
		or ((<LCFRSEdge>a.value).score == (<LCFRSEdge>b.value).score
		and a.count < b.count))


cdef class Agenda:
	"""
	Priority Queue based on binary heap which implements decrease-key and
	remove by marking entries as invalid. Provides dictionary-like interface.

	Can be initialized with an iterable; order of equivalent values remains and
	the best priorities are retained on duplicate keys. """
	def __init__(self, iterable=None):
		cdef Entry entry, oldentry
		self.counter = 1
		self.length = 0
		self.heap = []
		self.mapping = {}
		if iterable:
			for k, v in iterable:
				entry = new_Entry(k, v, self.counter)
				if k in self.mapping:
					oldentry = <Entry>self.mapping[k]
					if cmpfun(entry, oldentry):
						oldentry.count = INVALID
						self.mapping[k] = entry
				else:
					self.mapping[k] = entry
					self.counter += 1
				self.heap.append(entry)
			self.length = len(self.mapping)
			heapify(self.heap, cmpfun)

	cdef setitem(self, key, value):
		""" Like agenda[key] = value, but bypass Python API. """
		cdef Entry oldentry, entry
		if key in self.mapping:
			oldentry = <Entry>self.mapping[key]
			entry = <Entry>Entry.__new__(Entry)
			entry.key = key
			entry.value = value
			entry.count = oldentry.count
			self.mapping[key] = entry
			heappush(self.heap, entry, cmpfun)
			oldentry.count = INVALID
		else:
			self.counter += 1
			self.length += 1
			entry = <Entry>Entry.__new__(Entry)
			entry.key = key
			entry.value = value
			entry.count = self.counter
			self.mapping[key] = entry
			heappush(self.heap, entry, cmpfun)

	cdef setifbetter(self, key, value):
		""" Sets an item, but only if item is new or has lower score.
		Equivalent to:
		if if key not in agenda or val < agenda[key]:
			agenda[key] = val """
		cdef Entry oldentry
		if key in self.mapping:
			oldentry = <Entry>self.mapping[key]
			if value >= oldentry.value:
				return
		self.setitem(key, value)

	cdef getitem(self, key):
		""" Like agenda[key], but bypass Python API. """
		cdef Entry entry
		entry = <Entry>self.mapping[key]
		return entry.value

	cdef object replace(self, key, value):
		""" :returns: current value for key, and also change its value.

		Equivalent to vv = d[k]; d[k] = v; return vv """
		cdef Entry entry, oldentry = <Entry>self.mapping[key]
		entry = <Entry>Entry.__new__(Entry)
		entry.key = key
		entry.value = value
		entry.count = oldentry.count
		self.mapping[key] = entry
		heappush(self.heap, entry, cmpfun)
		oldentry.count = INVALID
		return oldentry.value

	cdef Entry peekentry(self):
		""" Get the current best entry, while keeping it on the agenda. """
		cdef Entry entry
		cdef Py_ssize_t n = PyList_GET_SIZE(self.heap)
		if n == 0:
			raise IndexError("peek at empty heap")
		entry = <Entry>(self.heap[0])
		while entry.count == 0:
			if n == 1:
				raise IndexError("peek at empty heap")
			#replace first element with last element
			self.heap[0] = self.heap.pop()
			#and restore heap invariant
			siftdown(self.heap, 0, cmpfun)
			n -= 1
			entry = <Entry>(self.heap[0])
		return entry

	cdef Entry popentry(self):
		""" like popitem(), but avoids tuple construction by returning an Entry
		object """
		cdef Entry entry = <Entry>heappop(self.heap, cmpfun)
		while not entry.count:
			entry = <Entry>heappop(self.heap, cmpfun)
		del self.mapping[entry.key]
		self.length -= 1
		return entry

	cdef bint contains(self, key):
		""" Like `key in agenda`, but bypass the Python API. """
		return key in self.mapping

	def peekitem(self):
		""" Get the current best (key, value) pair, while keeping it on the
		agenda. """
		cdef Entry entry = self.peekentry()
		return entry.key, entry.value

	# standard dict() methods
	def pop(self, key):
		""" :returns: value for agenda[key] and remove it. """
		cdef Entry entry
		if key is None:
			return self.popentry().value
		entry = <Entry>(self.mapping.pop(key))
		entry.count = INVALID
		self.length -= 1
		return entry.value

	def popitem(self):
		""" :returns: best scoring (key, value) pair; removed from agenda. """
		cdef Entry entry = self.popentry()
		return entry.key, entry.value

	def __delitem__(self, key):
		""" Remove key from heap. """
		(<Entry>self.mapping[key]).count = INVALID
		self.length -= 1
		del self.mapping[key]

	def update(self, *a, **kw):
		""" Change score of items given a sequence of (key, value) pairs. """
		for b in a:
			for k, v in b:
				self[k] = v
		for k, v in kw.items():
			self[k] = v

	def clear(self):
		""" Remove all items from agenda. """
		self.counter = 1
		del self.heap[:]
		self.mapping.clear()

	def __repr__(self):
		return '%s({%s})' % (self.__class__.__name__, ", ".join(
				['%r: %r' % ((<Entry>a).key, (<Entry>a).value)
				for a in self.heap if (<Entry>a).count]))

	def __contains__(self, key):
		return key in self.mapping

	def __getitem__(self, key):
		return self.getitem(key)

	def __setitem__(self, key, value):
		self.setitem(key, value)

	def __str__(self):
		return self.__repr__()

	def __iter__(self):
		return iter(self.mapping)

	def __len__(self):
		return self.length

	def __nonzero__(self):
		return self.length != 0

	def keys(self):
		""" :returns: keys in agenda. """
		return self.mapping.keys()

	def values(self):
		""" :returns: values in agenda. """
		return map(getval, self.mapping.values())

	def items(self):
		""" :returns: (key, value) pairs in agenda. """
		return zip(self.keys(), self.values())

cdef class EdgeAgenda:
	"""
	Priority Queue based on binary heap which implements decrease-key and
	remove by marking entries as invalid. Provides dictionary-like interface.

	Can be initialized with an iterable; order of equivalent values
	remains and the best priorities are retained on duplicate keys.

	This version is specialized to be used as agenda with edges. """
	def __init__(self, iterable=None):
		cdef Entry entry, oldentry
		self.counter = 1
		self.length = 0
		self.heap = []
		self.mapping = {}
		if iterable:
			for k, v in iterable:
				entry = new_Entry(k, v, self.counter)
				if k in self.mapping:
					oldentry = <Entry>self.mapping[k]
					if edgecmpfun(entry, oldentry):
						oldentry.count = INVALID
						self.mapping[k] = entry
				else:
					self.mapping[k] = entry
					self.counter += 1
				self.heap.append(entry)
			self.length = len(self.mapping)
			heapify(self.heap, edgecmpfun)

	cdef LCFRSEdge getitem(self, key):
		""" Like agenda[key], but bypass Python API. """
		cdef Entry entry
		entry = <Entry>self.mapping[key]
		return <LCFRSEdge>entry.value

	cdef setifbetter(self, key, value):
		""" sets an item, but only if item is new or has lower score """
		cdef Entry oldentry
		if key in self.mapping:
			oldentry = <Entry>self.mapping[key]
			if (<LCFRSEdge>value).score >= (<LCFRSEdge>oldentry.value).score:
				return
		self.setitem(key, <LCFRSEdge>value)

	cdef LCFRSEdge replace(self, key, value):
		""" return current value for key, and also change its value.
		equivalent to vv = d[k]; d[k] = v; return vv """
		cdef Entry entry, oldentry = <Entry>self.mapping[key]
		entry = <Entry>Entry.__new__(Entry)
		entry.key = key
		entry.value = value
		entry.count = oldentry.count
		self.mapping[key] = entry
		self.heap.append(entry)
		siftup(self.heap, 0, PyList_GET_SIZE(self.heap) - 1, edgecmpfun)
		oldentry.count = INVALID
		return <LCFRSEdge>oldentry.value

	# the following are identical except for `edgecmpfun`
	cdef Entry popentry(self):
		""" like popitem, but avoids tuple construction by returning an Entry
		object """
		cdef Entry entry = <Entry>heappop(self.heap, edgecmpfun)
		while not entry.count:
			entry = <Entry>heappop(self.heap, edgecmpfun)
		del self.mapping[entry.key]
		self.length -= 1
		return entry

	cdef Entry peekentry(self):
		""" Get the current best entry, while keeping it on the agenda. """
		cdef Entry entry
		cdef Py_ssize_t n = PyList_GET_SIZE(self.heap)
		if n == 0:
			raise IndexError("peek at empty heap")
		entry = <Entry>(self.heap[0])
		while entry.count == 0:
			if n == 1:
				raise IndexError("peek at empty heap")
			#replace first element with last element
			self.heap[0] = self.heap.pop()
			#and restore heap invariant
			siftdown(self.heap, 0, edgecmpfun)
			n -= 1
			entry = <Entry>(self.heap[0])
		return entry

	cdef setitem(self, key, value):
		""" Like agenda[key] = value, but bypass Python API. """
		cdef Entry oldentry, entry
		if key in self.mapping:
			oldentry = <Entry>self.mapping[key]
			entry = <Entry>Entry.__new__(Entry)
			entry.key = key
			entry.value = value
			entry.count = oldentry.count
			self.mapping[key] = entry
			self.heap.append(entry)
			siftup(self.heap, 0, PyList_GET_SIZE(self.heap) - 1, edgecmpfun)
			oldentry.count = INVALID
		else:
			self.counter += 1
			self.length += 1
			entry = <Entry>Entry.__new__(Entry)
			entry.key = key
			entry.value = value
			entry.count = self.counter
			self.mapping[key] = entry
			self.heap.append(entry)
			siftup(self.heap, 0, PyList_GET_SIZE(self.heap) - 1, edgecmpfun)

	# identical to Agenda() methods
	cdef bint contains(self, key):
		""" Like `key in agenda`, but bypass the Python API. """
		return key in self.mapping

	def pop(self, key):
		""" :returns: value for agenda[key] and remove it. """
		cdef Entry entry
		if key is None:
			return self.popentry().value
		entry = <Entry>(self.mapping.pop(key))
		entry.count = INVALID
		self.length -= 1
		return entry.value

	def popitem(self):
		""" :returns: best scoring (key, value) pair, removed from agenda. """
		cdef Entry entry = self.popentry()
		return entry.key, entry.value

	def __delitem__(self, key):
		(<Entry>self.mapping[key]).count = INVALID
		self.length -= 1
		del self.mapping[key]

	def update(self, *a, **kw):
		""" Change score of items given a sequence of (key, value) pairs. """
		for b in a:
			for k, v in b:
				self[k] = v
		for k, v in kw.items():
			self[k] = v

	def clear(self):
		""" Remove all items from agenda. """
		self.counter = 1
		del self.heap[:]
		self.mapping.clear()

	def __repr__(self):
		return '%s({%s})' % (self.__class__.__name__, ", ".join(
				['%r: %r' % ((<Entry>a).key, (<Entry>a).value)
				for a in self.heap if (<Entry>a).count]))

	def __contains__(self, key):
		return key in self.mapping

	def __getitem__(self, key):
		return self.getitem(key)

	def __setitem__(self, key, value):
		self.setitem(key, value)

	def __str__(self):
		return self.__repr__()

	def __iter__(self):
		return iter(self.mapping)

	def __len__(self):
		return self.length

	def __nonzero__(self):
		return self.length != 0

	def keys(self):
		""" :returns: keys in agenda. """
		return self.mapping.keys()

	def values(self):
		""" :returns: values in agenda. """
		return map(getval, self.mapping.values())

	def items(self):
		""" :returns: (key, value) pairs in agenda. """
		return zip(self.keys(), self.values())


#a more efficient nsmallest implementation. Assumes items are Edge objects.
cdef list nsmallest(int n, object iterable):
	""" return an _unsorted_ list of the n best items in a list """
	cdef list items = list(iterable)
	if len(items) > 1:
		quickfindfirstk(items, 0, len(items) - 1, n)
	return items[:n]


cdef inline void quickfindfirstk(list items, int left, int right, int k):
	""" quicksort k-best selection """
	# select pivot index between left and right
	# middle between left & right
	cdef int pivot = left + (right - left) // 2
	cdef int pivotnewindex = partition(items, left, right, pivot)
	if pivotnewindex > k:
		if pivotnewindex - 1 > left:
			# new condition
			quickfindfirstk(items, left, pivotnewindex - 1, k)
	elif pivotnewindex < k:
		if right > pivotnewindex + 1:
			quickfindfirstk(items, pivotnewindex + 1, right, k)


cdef inline int partition(list items, int left, int right, int pivot):
	cdef Edge pivotvalue = <Edge>(items[pivot])
	# Move pivot to end
	items[pivot], items[right] = items[right], items[pivot]
	cdef int i, storeindex = left
	for i in range(left, right):
		if (<Edge>items[i]).inside < pivotvalue.inside:
			items[i], items[storeindex] = items[storeindex], items[i]
			storeindex += 1
	# Move pivot to its final place
	items[storeindex], items[right] = items[right], items[storeindex]
	return storeindex


# heap operations (adapted from heapq)
cdef inline Entry heappop(list heap, CmpFun cmpfun):
	cdef Py_ssize_t n = PyList_GET_SIZE(heap)
	cdef Entry entry
	if n == 0:
		raise IndexError("pop from empty heap")
	elif n == 1:
		entry = <Entry>heap.pop()
	else:
		#replace first element with last element and restore heap invariant
		entry = <Entry>(PyList_GET_ITEM(heap, 0))
		heap[0] = heap.pop()
		siftdown(heap, 0, cmpfun)
	return entry


cdef inline void heappush(list heap, Entry entry, CmpFun cmpfun):
	# place at the end and swap with parents until heap invariant holds
	heap.append(entry)
	siftup(heap, 0, PyList_GET_SIZE(heap) - 1, cmpfun)


cdef inline void heapify(list heap, CmpFun cmpfun):
	""" Transform list into a heap, in-place, in O(len(heap)) time. """
	cdef int i
	for i in range(PyList_GET_SIZE(heap) // HEAP_ARITY, -1, -1):
		siftdown(heap, i, cmpfun)


# shifts only apply for binary tree
cdef inline int _parent(int i):
	return (i - 1) // HEAP_ARITY
	#return (i - 1) >> 1


cdef inline int _left(int i):
	return i * HEAP_ARITY + 1
	#return (i << 1) + 1


cdef inline int _right(int i):
	""" for documentation purposes; not used. """
	return i * HEAP_ARITY + 2
	#return (i + 1) << 1


cdef inline void siftdown(list heap, int pos, CmpFun cmpfun):
	cdef int startpos = pos, childpos = _left(pos), rightpos
	cdef int endpos = PyList_GET_SIZE(heap)
	cdef Entry newitem = <Entry>PyList_GET_ITEM(heap, pos)
	while childpos < endpos:
		for rightpos in range(childpos + 1, childpos + HEAP_ARITY):
			if (rightpos < endpos and
				cmpfun(<Entry>(PyList_GET_ITEM(heap, rightpos)),
					<Entry>(PyList_GET_ITEM(heap, childpos)))):
				childpos = rightpos
		heap[pos] = <Entry>PyList_GET_ITEM(heap, childpos)
		pos = childpos
		childpos = _left(pos)
	heap[pos] = newitem
	siftup(heap, startpos, pos, cmpfun)


cdef inline void siftup(list heap, int startpos, int pos, CmpFun cmpfun):
	cdef int parentpos
	cdef Entry parent, newitem = <Entry>PyList_GET_ITEM(heap, pos)
	while pos > startpos:
		parentpos = _parent(pos)
		parent = <Entry>PyList_GET_ITEM(heap, parentpos)
		if cmpfun(parent, newitem):
			break
		heap[pos] = parent
		pos = parentpos
	heap[pos] = newitem


#==============================================================================
from unittest import TestCase
testN = 1000


def getval(Entry entry):
	return entry.value


class TestHeap(TestCase):
	def check_invariants(self, Agenda h):
		for i in range(len(h)):
			if i > 0:
				self.assertTrue(getval(h.heap[_parent(i)]) <= getval(h.heap[i]))

	def make_data(self):
		from random import random
		pairs = [(random(), random()) for i in range(testN)]
		h = Agenda()
		d = {}
		for k, v in pairs:
			h[k] = v
			d[k] = v

		pairs.sort(key=itemgetter(1), reverse=True)
		return h, pairs, d

	def test_contains(self):
		h, pairs, d = self.make_data()
		h, pairs2, d = self.make_data()
		for k, v in pairs + pairs2:
			self.assertEqual(k in h, k in d)

	def test_len(self):
		h, pairs, d = self.make_data()
		self.assertEqual(len(h), len(d))

	def test_popitem(self):
		h, pairs, d = self.make_data()
		while pairs:
			v = h.popitem()
			v2 = pairs.pop(-1)
			self.assertEqual(v, v2)
			d.pop(v[0])
			self.assertEqual(len(h), len(d))
			self.assertEqual(set(h.items()), set(d.items()))
		self.assertEqual(len(h), 0)

	def test_popitem_ties(self):
		h = Agenda()
		for i in range(testN):
			h[i] = 0.
		for i in range(testN):
			k, v = h.popitem()
			self.assertEqual(v, 0.)
			self.check_invariants(h)

	def test_popitem_ties_fifo(self):
		h = Agenda()
		for i in range(testN):
			h[i] = 0.
		for i in range(testN):
			k, v = h.popitem()
			self.assertEqual(k, i)
			self.assertEqual(v, 0.)
			self.check_invariants(h)

	def test_peek(self):
		h, pairs, d = self.make_data()
		while pairs:
			v = h.peekitem()[0]
			h.popitem()
			v2 = pairs.pop(-1)
			self.assertEqual(v, v2[0])
		self.assertEqual(len(h), 0)

	def test_iter(self):
		h, pairs, d = self.make_data()
		self.assertEqual(list(h), list(d))

	def test_keys(self):
		h, pairs, d = self.make_data()
		self.assertEqual(sorted(h.keys()), sorted(d.keys()))

	def test_values(self):
		h, pairs, d = self.make_data()
		self.assertEqual(sorted(h.values()), sorted(d.values()))

	def test_items(self):
		h, pairs, d = self.make_data()
		self.assertEqual(sorted(h.items()), sorted(d.items()))

	def test_del(self):
		h, pairs, d = self.make_data()
		while pairs:
			k, v = pairs.pop(len(pairs) // 2)
			del h[k]
			del d[k]
			self.assertEqual(len(h), len(d))
			self.assertEqual(set(h.items()), set(d.items()))
		self.assertEqual(len(h), 0)

	def test_pop(self):
		h, pairs, d = self.make_data()
		while pairs:
			k, v = pairs.pop(-1)
			v2 = h.pop(k)
			v3 = d.pop(k)
			self.assertEqual(v, v2)
			self.assertEqual(v, v3)
			self.assertEqual(len(h), len(d))
			self.assertEqual(set(h.items()), set(d.items()))
		self.assertEqual(len(h), 0)

	def test_change(self):
		h, pairs, d = self.make_data()
		k, v = pairs[testN // 2]
		h[k] = 0.5
		pairs[testN // 2] = (k, 0.5)
		pairs.sort(key=itemgetter(1), reverse=True)
		while pairs:
			v = h.popitem()
			v2 = pairs.pop()
			self.assertEqual(v, v2)
		self.assertEqual(len(h), 0)

	def test_init(self):
		h, pairs, d = self.make_data()
		h = Agenda(d.items())
		while pairs:
			v = h.popitem()
			v2 = pairs.pop()
			self.assertEqual(v, v2)
			d.pop(v[0])
		self.assertEqual(len(h), len(d))
		self.assertEqual(len(h), 0)

	def test_repr(self):
		h, pairs, d = self.make_data()
		#self.assertEqual(h, eval(repr(h)))
		tmp = repr(h)  # 'Agenda({....})'
		#strip off class name
		dstr = tmp[tmp.index('(') + 1:tmp.rindex(')')]
		self.assertEqual(d, eval(dstr))


def test(verbose=False):
	import sys
	if sys.version[0] >= '3':
		import test.support as test_support  # Python 3
	else:
		import test.test_support as test_support  # Python 2
	test_support.run_unittest(TestHeap)

	# verify reference counting
	if verbose and hasattr(sys, "gettotalrefcount"):
		import gc
		counts = [None] * 5
		for i in xrange(len(counts)):
			test_support.run_unittest(TestHeap)
			gc.collect()
			counts[i] = sys.gettotalrefcount()
		print(counts)



""" Implementation of Sangati et al. (2010), Efficiently extract recurring tree
fragments from large treebanks.
Moschitti (2006): Making Tree Kernels practical for Natural Language
Learning. """

from __future__ import print_function
import re
import io
from collections import defaultdict, Counter as multiset
from functools import partial
from array import array
from discodop.tree import Tree
from discodop.grammar import lcfrs_productions
from discodop.treetransforms import binarize, introducepreterminals

cimport cython
from libc.stdlib cimport malloc, free
from cpython.array cimport array, clone
from discodop._grammar cimport ULong, UInt
from discodop.containers cimport Node, NodeArray, Ctrees
from discodop.bit cimport iteratesetbits, abitcount, subset, ulongcpy, \
		ulongset, setunioninplace

cdef extern from "macros.h":
	int BITNSLOTS(int nb)
	void SETBIT(ULong a[], int b)
	void CLEARBIT(ULong a[], int b)
	ULong TESTBIT(ULong a[], int b)
	int IDX(int i, int j, int jmax, int kmax)

# template to create arrays of this type.
cdef array uintarray = array('I', ())

# NB: (?: ) is a non-grouping operator; the second ':' is part of what we match
FRONTIERORTERMRE = re.compile(r' ([0-9]+)(?::[0-9]+)?\b')  # all leaf nodes
TERMINDICESRE = re.compile(r'\([^(]+ ([0-9]+)\)')  # leaf nodes w/term. indices
FRONTIERRE = re.compile(r' ([0-9]+):([0-9]+)\b')  # non-terminal frontiers
LABEL = re.compile(' *\( *([^ ()]+) *')


# bitsets representing fragments are ULong arrays with the number of elements
# determined by SLOTS. While SLOTS is not a constant nor a global, it is
# set just once for a treebank to fit its largest tree.
# After SLOTS elements, this struct follows:
cdef packed struct BitsetTail:
	UInt id  # tree no. of which this fragment was extracted
	short root  # index of the root of the fragment in the NodeArray


# we wrap bitsets in bytes objects, because these can be (1) used in
# dictionaries and (2) passed between processes.
# we leave one byte for NUL-termination:
# data (SLOTS * 8), id (4), root (2), NUL (1)
cdef inline bytes wrap(ULong *data, short SLOTS):
	""" Wrap bitset in bytes object for handling in Python space. """
	return (<char *>data)[:SLOTS * sizeof(ULong) + sizeof(BitsetTail)]


# use getters & setters because a cdef class would incur overhead & indirection
# of a python object, and with a struct the root & id fields must be in front
# which seems to lead to bad hashing behavior (?)
cdef inline ULong *getpointer(object wrapper):
	""" Get pointer to bitset from wrapper. """
	return <ULong *><char *><bytes>wrapper


cdef inline UInt getid(ULong *data, short SLOTS):
	""" Get id of fragment in a bitset. """
	return (<BitsetTail *>&data[SLOTS]).id


cdef inline short getroot(ULong *data, short SLOTS):
	""" Get root of fragment in a bitset. """
	return (<BitsetTail *>&data[SLOTS]).root


cdef inline void setrootid(ULong *data, short root, UInt id, short SLOTS):
	""" Set root and id of fragment in a bitset. """
	cdef BitsetTail *tail = <BitsetTail *>&data[SLOTS]
	tail.id = id
	tail.root = root


cpdef fastextractfragments(Ctrees trees1, list sents1, int offset, int end,
		list labels, Ctrees trees2=None, list sents2=None, bint approx=True,
		bint debug=False, bint discontinuous=False, bint complement=False):
	""" Seeks the largest fragments in treebank(s) with a linear time tree
	kernel

	- scenario 1: recurring fragments in single treebank, use:
		fastextractfragments(trees1, sents1, offset, end, labels)
	- scenario 2: common fragments in two treebanks:
		fastextractfragments(trees1, sents1, offset, end, trees2, sents2)

	offset and end can be used to divide the work over multiple processes.
	they are indices of trees1 to work on (default is all).
	when debug is enabled a contingency table is shown for each pair of trees.
	when complement is true, the complement of the recurring fragments in each
	pair of trees is extracted as well. """
	cdef:
		int n, m, start = 0, end2
		short SLOTS  # the number of bitsets needed to cover the largest tree
		ULong *CST = NULL  # Common Subtree Table
		ULong *scratch, *bitset  # temporary variables
		NodeArray a, b, *ctrees1, *ctrees2
		Node *anodes, *bnodes
		list asent
		dict fragments = {}
		set inter = set()
	if approx:
		fragments = <dict>defaultdict(one)

	if trees2 is None:
		trees2 = trees1
	ctrees1 = trees1.trees
	ctrees2 = trees2.trees
	SLOTS = BITNSLOTS(max(trees1.maxnodes, trees2.maxnodes) + 1)
	CST = <ULong *>malloc(trees2.maxnodes * SLOTS * sizeof(ULong))
	scratch = <ULong *>malloc((SLOTS + 2) * sizeof(ULong))
	assert CST is not NULL and scratch is not NULL
	end2 = trees2.len
	# find recurring fragments
	for n in range(offset, min(end or trees1.len, trees1.len)):
		a = ctrees1[n]
		asent = <list>(sents1[n])
		anodes = &trees1.nodes[a.offset]
		if sents2 is None:
			start = n + 1
		for m in range(start, end2):
			b = ctrees2[m]
			bnodes = &trees2.nodes[b.offset]
			# initialize table
			ulongset(CST, 0UL, b.len * SLOTS)
			# fill table
			fasttreekernel(anodes, bnodes, a.len, b.len, CST, SLOTS)
			# dump table
			if debug:
				print(n, m)
				dumpCST(CST, a, b, anodes, bnodes, asent, (sents2 or sents1)[m],
					labels, SLOTS, True)
			# extract results
			extractbitsets(CST, anodes, bnodes, b.root, n, inter,
					scratch, SLOTS)
			# extract complementary fragments?
			if complement:
				# combine bitsets of inter together with bitwise or
				ulongset(scratch, 0UL, SLOTS)
				for wrapper in inter:
					setunioninplace(scratch, getpointer(wrapper), SLOTS)
				# extract bitsets in A from result, without regard for B
				extractcompbitsets(scratch, anodes, a.root, n, inter,
						SLOTS, NULL)
		# collect string representations of fragments
		for wrapper in inter:
			bitset = getpointer(wrapper)
			if discontinuous:
				frag = getsubtree(anodes, bitset, labels,
						getroot(bitset, SLOTS))
				frag = getsent(frag, asent)
			else:
				frag = getsubtreeunicode(anodes, bitset, labels, asent,
						getroot(bitset, SLOTS))
			if approx:
				fragments[frag] += 1
			elif frag not in fragments:
				fragments[frag] = wrapper
		inter.clear()
	free(CST)
	free(scratch)
	return fragments


cdef inline void fasttreekernel(Node *a, Node *b, int alen, int blen,
		ULong *CST, short SLOTS):
	""" Fast Tree Kernel (average case linear time). Expects trees to be sorted
	according to their productions (in descending order, with terminals as -1).
	Moschitti (2006): Making Tree Kernels practical for Natural Language
	Learning. """
	# i as an index to a, j to b, and jj is a temp index starting at j.
	cdef int i = 0, j = 0, jj = 0
	while i < alen and j < blen:
		if a[i].prod < b[j].prod:
			i += 1
		elif a[i].prod > b[j].prod:
			j += 1
		else:
			while i < alen and a[i].prod == b[j].prod:
				jj = j
				while jj < blen and a[i].prod == b[jj].prod:
					SETBIT(&CST[jj * SLOTS], i)
					jj += 1
				i += 1


cdef inline extractbitsets(ULong *CST, Node *a, Node *b,
		short j, int n, set results, ULong *scratch, short SLOTS):
	""" Visit nodes of 'b' in pre-order traversal. store bitsets of connected
	subsets of 'a' as they are encountered, following the common nodes
	specified in CST. j is the node in 'b' to start with, which changes with
	each recursive step. 'n' is the identifier of this tree which is stored
	with extracted fragments. """
	cdef ULong *bitset = &CST[j * SLOTS]
	cdef ULong cur = bitset[0]
	cdef short idx = 0
	cdef short i = iteratesetbits(bitset, SLOTS, &cur, &idx)
	while i != -1:
		ulongset(scratch, 0UL, SLOTS)
		extractat(CST, scratch, a, b, i, j, SLOTS)
		setrootid(scratch, i, n, SLOTS)
		results.add(wrap(scratch, SLOTS))
		i = iteratesetbits(bitset, SLOTS, &cur, &idx)
	if b[j].left >= 0:
		extractbitsets(CST, a, b, b[j].left, n, results, scratch, SLOTS)
		if b[j].right >= 0:
			extractbitsets(CST, a, b, b[j].right, n, results, scratch, SLOTS)


cdef inline void extractat(ULong *CST, ULong *result, Node *a, Node *b,
		short i, short j, short SLOTS):
	""" Traverse tree a and b in parallel to extract a connected subset """
	SETBIT(result, i)
	CLEARBIT(&CST[j * SLOTS], i)
	if a[i].left < 0:
		return
	elif TESTBIT(&CST[b[j].left * SLOTS], a[i].left):
		extractat(CST, result, a, b, a[i].left, b[j].left, SLOTS)
	if a[i].right < 0:
		return
	elif TESTBIT(&CST[b[j].right * SLOTS], a[i].right):
		extractat(CST, result, a, b, a[i].right, b[j].right, SLOTS)


cpdef exactcounts(Ctrees trees1, Ctrees trees2, list bitsets,
		bint fast=True, bint indices=False):
	""" Given a set of fragments from trees1 as bitsets, find all occurences
	of those fragments in trees2 (which may be equal to trees1).
	By default, exact counts are collected. When indices is True, a multiset
	of indices is returned for each fragment; the indices point to the trees
	where the fragments (described by the bitsets) occur. This is useful to
	look up the contexts of fragments, or when the order of sentences has
	significance which makes it possible to interpret the set of indices as
	time-series data.
	The reason we need to do this separately from extracting maximal fragments
	is that a fragment can occur in other trees where it was not a maximal. """
	cdef:
		array counts
		list theindices
		set candidates
		short i, j, SLOTS = BITNSLOTS(max(trees1.maxnodes, trees2.maxnodes) + 1)
		UInt n, m, *countsp = NULL
		ULong *bitset
		NodeArray a, b
		Node *anodes, *bnodes
		ULong cur
		short idx
	if indices:
		theindices = [multiset() for _ in bitsets]
	else:
		counts = clone(uintarray, len(bitsets), True)
		countsp = counts.data.as_uints
	# compare one bitset to each tree for each unique fragment.
	for n, wrapper in enumerate(bitsets):
		bitset = getpointer(wrapper)
		a = trees1.trees[getid(bitset, SLOTS)]  # fragment came from this tree
		anodes = &trees1.nodes[a.offset]
		idx = 0
		cur = bitset[idx]
		i = iteratesetbits(bitset, SLOTS, &cur, &idx)
		assert i != -1
		candidates = <set>(trees2.treeswithprod[anodes[i].prod]).copy()
		while True:
			i = iteratesetbits(bitset, SLOTS, &cur, &idx)
			if i == -1 or i >= a.len:  # FIXME. why is 2nd condition necessary?
				break
			candidates &= <set>(trees2.treeswithprod[anodes[i].prod])
		i = getroot(bitset, SLOTS)  # root of fragment in tree 'a'
		if indices:
			matches = theindices[n]
		for m in candidates:
			b = trees2.trees[m]
			bnodes = &trees2.nodes[b.offset]
			for j in range(b.len):
				if anodes[i].prod == bnodes[j].prod:
					if containsbitset(anodes, bnodes, bitset, i, j):
						if indices:
							matches[m] += 1
						else:
							countsp[n] += 1
				elif fast and anodes[i].prod < bnodes[j].prod:
					break
	return theindices if indices else counts


cdef inline int containsbitset(Node *a, Node *b, ULong *bitset,
		short i, short j):
	""" Recursively check whether fragment starting from a[i] described by
	bitset is equal to b[j], i.e., whether b contains that fragment from a. """
	if a[i].prod != b[j].prod:
		return 0
	elif a[i].left < 0:
		return 1
	elif TESTBIT(bitset, a[i].left):
		if not containsbitset(a, b, bitset, a[i].left, b[j].left):
			return 0
	if a[i].right < 0:
		return 1
	elif TESTBIT(bitset, a[i].right):
		return containsbitset(a, b, bitset, a[i].right, b[j].right)
	return 1


cpdef dict coverbitsets(Ctrees trees, list sents, list labels,
		short maxnodes, bint discontinuous):
	""" Utility function to generate one bitset for each type of production. """
	cdef:
		dict result = {}
		int p, i, n = -1
		short SLOTS = BITNSLOTS(maxnodes + 1)
		ULong *scratch = <ULong *>malloc((SLOTS + 2) * sizeof(ULong))
		Node *nodes
	assert scratch is not NULL
	assert SLOTS, SLOTS
	for p, treeindices in enumerate(trees.treeswithprod):
		ulongset(scratch, 0UL, SLOTS)
		# slightly convoluted way of getting an arbitrary set member:
		n = next(iter(treeindices))
		nodes = &trees.nodes[trees.trees[n].offset]
		for i in range(trees.trees[n].len):
			if nodes[i].prod == p:
				SETBIT(scratch, i)
				setrootid(scratch, i, n, SLOTS)
				break
		else:
			raise ValueError("production not found. wrong index?")
		if discontinuous:
			frag = getsubtree(nodes, scratch, labels, i)
			frag = getsent(frag, sents[n])
		else:
			frag = getsubtreeunicode(nodes, scratch, labels,
					sents[n], i)
		result[frag] = wrap(scratch, SLOTS)
	return result


cpdef dict completebitsets(Ctrees trees, list sents, list labels,
		bint discontinuous=False):
	""" Utility function to generate bitsets corresponding to whole trees
	in the input. """
	cdef:
		dict result = {}
		list sent
		int n, i
		short SLOTS = BITNSLOTS(trees.maxnodes + 1)
		ULong *scratch = <ULong *>malloc((SLOTS + 2) * sizeof(ULong))
		Node *nodes
	for n in range(trees.len):
		ulongset(scratch, 0UL, SLOTS)
		nodes = &trees.nodes[trees.trees[n].offset]
		sent = sents[n]
		for i in range(trees.trees[n].len):
			if nodes[i].left >= 0 or sent[termidx(nodes[i].left)] is not None:
				SETBIT(scratch, i)
		setrootid(scratch, trees.trees[n].root, n, SLOTS)
		if discontinuous:
			frag = (strtree(nodes, labels, trees.trees[n].root), sent)
		else:
			frag = unicodetree(nodes, labels, sent, trees.trees[n].root)
		result[frag] = wrap(scratch, SLOTS)
	return result


cdef inline void extractcompbitsets(ULong *bitset, Node *a,
		int i, int n, set results, short SLOTS, ULong *scratch):
	""" Visit nodes of 'a' in pre-order traversal. store bitsets of connected
	subsets of 'a' as they are encountered, following the nodes specified in
	the complement of 'bitset'. i is the node in 'a' to start with, which
	changes with each recursive step. 'n' is the identifier of this tree which
	is stored with extracted fragments. """
	cdef bint start = scratch is NULL and not TESTBIT(bitset, i)
	if start:  # start extracting a fragment
		# need to create a new array because further down in the recursion
		# other fragments may be encountered which should not overwrite
		# this one
		scratch = <ULong *>malloc((SLOTS + 2) * sizeof(ULong))
		assert scratch is not NULL
		setrootid(scratch, i, n, SLOTS)
	elif TESTBIT(bitset, i):  # stop extracting for this fragment
		scratch = NULL
	if scratch is not NULL:
		SETBIT(scratch, i)
	if a[i].left >= 0:
		extractcompbitsets(bitset, a, a[i].left, n, results, SLOTS, scratch)
		if a[i].right >= 0:
			extractcompbitsets(bitset, a, a[i].right, n, results, SLOTS, scratch)
	if start:  # store fragment
		results.add(wrap(scratch, SLOTS))
		free(scratch)


cpdef extractfragments(Ctrees trees1, list sents1, int offset, int end,
		list labels, Ctrees trees2=None, list sents2=None, bint approx=True,
		bint debug=False, bint discontinuous=False):
	""" Seeks the largest fragments in treebank(s)

	- scenario 1: recurring fragments in single treebank, use:
		extractfragments(trees1, sents1, offset, end)
	- scenario 2: common fragments in two treebanks:
		extractfragments(trees1, sents1, offset, end, trees2, sents2)

	offset and end can be used to divide the work over multiple processes.
	offset is the starting point in trees1, end is the number of trees from
	trees1 to work on. When debug is enabled a contingency table is shown for
	each pair of trees. """
	cdef:
		int n, m, aa, bb, start = 0
		short SLOTS
		ULong *CST, *scratch, *bitset
		NodeArray *ctrees1, *ctrees2, a, b
		Node *anodes, *bnodes
		list asent
		dict fragments = {}
		set inter = set()

	if approx:
		fragments = <dict>defaultdict(one)
	if trees2 is None:
		trees2 = trees1
	ctrees1 = trees1.trees
	ctrees2 = trees2.trees
	SLOTS = BITNSLOTS(max(trees1.maxnodes, trees2.maxnodes) + 1)
	CST = <ULong *>malloc(trees1.maxnodes * trees2.maxnodes
		* SLOTS * sizeof(ULong))
	scratch = <ULong *>malloc((SLOTS + 2) * sizeof(ULong))
	assert CST is not NULL and scratch is not NULL

	# find recurring fragments
	for n in range(offset, (end or trees1.len)):
		a = ctrees1[n]
		anodes = &trees1.nodes[a.offset]
		asent = <list>(sents1[n])
		if sents2 is None:
			start = n + 1
		for m in range(start, trees2.len):
			b = ctrees2[m]
			bnodes = &trees2.nodes[b.offset]
			# initialize table
			ulongset(CST, 0UL, a.len * b.len * SLOTS)
			# fill table
			for aa in range(a.len):
				for bb in range(b.len):
					if not TESTBIT(&CST[IDX(aa, bb, b.len, SLOTS)], a.len):
						getCST(anodes, bnodes, a.len, b.len, CST, aa, bb, SLOTS)
			# dump table
			if debug:
				print(n, m)
				dumpCST(CST, a, b, anodes, bnodes, asent,
					(sents2[m] if sents2 else sents1[m]), labels, SLOTS)
			# extract results
			inter.update(getnodeset(CST, a.len, b.len, n, scratch, SLOTS))
		for wrapper in inter:
			bitset = getpointer(wrapper)
			if discontinuous:
				frag = getsubtree(anodes, bitset, labels, getroot(bitset, SLOTS))
				frag = getsent(frag, asent)
			else:
				frag = getsubtreeunicode(anodes, bitset, labels, asent,
						getroot(bitset, SLOTS))
			if approx:
				fragments[frag] += 1
			else:
				fragments[frag] = wrapper
		inter.clear()
	free(CST)
	return fragments


cdef inline void getCST(Node *a, Node *b, int alen, int blen, ULong *CST,
		int i, int j, short SLOTS):
	""" Recursively build common subtree table (CST) for subtrees a[i], b[j]. """
	cdef ULong *child, *bitset = &CST[IDX(i, j, blen, SLOTS)]
	SETBIT(bitset, alen)  # mark cell as visited
	# compare label & arity / terminal; assume presence of arity markers.
	if a[i].prod == b[j].prod:
		SETBIT(bitset, i)
		# lexical production, no recursion
		if a[i].left < 0:
			return
		# normal production, recurse or use cached value
		child = &CST[IDX(a[i].left, b[j].left, blen, SLOTS)]
		if not TESTBIT(child, alen):
			getCST(a, b, alen, blen, CST, a[i].left, b[j].left, SLOTS)
		for n in range(SLOTS):
			bitset[n] |= child[n]
		if a[i].right < 0:  # and b[j].right < 0: # unary node
			return
		child = &CST[IDX(a[i].right, b[j].right, blen, SLOTS)]
		if not TESTBIT(child, alen):
			getCST(a, b, alen, blen, CST, a[i].right, b[j].right, SLOTS)
		for n in range(SLOTS):
			bitset[n] |= child[n]


cdef inline set getnodeset(ULong *CST, int alen, int blen, int n,
		ULong *scratch, short SLOTS):
	""" Extract the largest, connected bitsets from the common subtree
	table (CST). """
	cdef:
		ULong *bitset
		int i, j
		set finalnodeset = set()
	for i in range(alen):
		for j in range(blen):
			bitset = &CST[IDX(i, j, blen, SLOTS)]
			if not TESTBIT(bitset, alen) or abitcount(bitset, SLOTS) < 2:
				continue
			ulongcpy(scratch, bitset, SLOTS)
			CLEARBIT(scratch, alen)
			setrootid(scratch, i, n, SLOTS)
			tmp = wrap(scratch, SLOTS)
			if tmp in finalnodeset:
				continue
			for bs in finalnodeset:
				if subset(scratch, getpointer(bs), SLOTS):
					break
				elif subset(getpointer(bs), scratch, SLOTS):
					finalnodeset.remove(bs)
					finalnodeset.add(tmp)
					break
			else:
				# completely new (disjunct) bitset
				finalnodeset.add(tmp)
	return finalnodeset


cdef inline short termidx(x):
	""" Translate representation for terminal indices. """
	return -x - 1


cdef inline str strtree(Node *tree, list labels, int i):
	""" produce string representation of (complete) tree, with indices as
	terminals. """
	if tree[i].left < 0:
		return "(%s %d)" % (labels[tree[i].prod], termidx(tree[i].left))
	elif tree[i].right < 0:
		return "(%s %s)" % (labels[tree[i].prod],
				strtree(tree, labels, tree[i].left))
	return "(%s %s %s)" % (labels[tree[i].prod],
			strtree(tree, labels, tree[i].left),
			strtree(tree, labels, tree[i].right))


cdef inline unicode unicodetree(Node *tree, list labels, list sent, int i):
	""" produce string representation of (complete) tree, with words as
	terminals. """
	if tree[i].left < 0:
		return u"(%s %s)" % (labels[tree[i].prod],
				(u"%d" % termidx(tree[i].left)
				if sent is None else
				sent[termidx(tree[i].left)] or u''))
	elif tree[i].right < 0:
		return u"(%s %s)" % (labels[tree[i].prod],
				unicodetree(tree, labels, sent, tree[i].left))
	return u"(%s %s %s)" % (labels[tree[i].prod],
			unicodetree(tree, labels, sent, tree[i].left),
			unicodetree(tree, labels, sent, tree[i].right))


cdef inline str getsubtree(Node *tree, ULong *bitset, list labels, int i):
	""" Turn bitset into string representation of tree,
	with indices as terminals. """
	if TESTBIT(bitset, i):
		if tree[i].left < 0:
			return "(%s %d)" % (labels[tree[i].prod], termidx(tree[i].left))
		elif tree[i].right < 0:
			return "(%s %s)" % (labels[tree[i].prod],
					getsubtree(tree, bitset, labels, tree[i].left))
		return "(%s %s %s)" % (labels[tree[i].prod],
				getsubtree(tree, bitset, labels, tree[i].left),
				getsubtree(tree, bitset, labels, tree[i].right))
	# node not in bitset, frontier non-terminal
	return "(%s %s)" % (labels[tree[i].prod], yieldranges(tree, i))


cdef inline unicode getsubtreeunicode(Node *tree, ULong *bitset, list labels,
		list sent, int i):
	""" Turn bitset into string representation of tree,
	with words as terminals. """
	if TESTBIT(bitset, i):
		if tree[i].left < 0:
			return u"(%s %s)" % (labels[tree[i].prod],
				(u"%d" % termidx(tree[i].left) if sent is None
				else sent[termidx(tree[i].left)]))
		elif tree[i].right < 0:
			return u"(%s %s)" % (labels[tree[i].prod],
				getsubtreeunicode(tree, bitset, labels, sent, tree[i].left))
		return u"(%s %s %s)" % (labels[tree[i].prod],
			getsubtreeunicode(tree, bitset, labels, sent, tree[i].left),
			getsubtreeunicode(tree, bitset, labels, sent, tree[i].right))
	# node not in bitset, frontier non-terminal
	return u"(%s )" % labels[tree[i].prod]


cdef inline yieldranges(Node *tree, int i):
	""" For discontinuous trees, return a string with the intervals of indices
	corresponding to the components in the yield of a node.
	The intervals are of the form start:end, where `end` is part of the
	interval. e.g., "0:1 2:4" corresponds to (0, 1) and (2, 3, 4). """
	cdef list yields = [], leaves = sorted(getyield(tree, i))
	cdef int a, start = -2, prev = -2
	for a in leaves:
		if a - 1 != prev:
			if prev != -2:
				yields.append("%d:%d" % (start, prev))
			start = a
		prev = a
	yields.append("%d:%d" % (start, prev))
	return ' '.join(yields)


cdef inline list getyield(Node *tree, int i):
	""" Recursively collect indices of terminals under a node. """
	if tree[i].left < 0:
		return [termidx(tree[i].left)]
	elif tree[i].right < 0:
		return getyield(tree, tree[i].left)
	return getyield(tree, tree[i].left) + getyield(tree, tree[i].right)


def repl(d):
	def f(x):
		return d[int(x.group(1))]
	return f


cdef getsent(frag, list sent):
	""" Select the words that occur in the fragment and renumber terminals in
	fragment such that the first index is 0 and any gaps have a width of 1.
	Expects a tree as string where frontier nodes are marked with intervals.

	>>> getsent('(S (NP 2) (VP 4))', ['The', 'tall', 'man', 'there', 'walks'])
	('(S (NP 0) (VP 2))', ('man', None, 'walks'))
	>>> getsent('(VP (VB 0) (PRT 3))', ['Wake', 'your', 'friend', 'up'])
	('(VP (VB 0) (PRT 2))', ('Wake', None, 'up'))
	>>> getsent('(S (NP 2:2 4:4) (VP 1:1 3:3))',['Walks','the','quickly','man'])
	('(S (NP 1 3) (VP 0 2))', (None, None, None, None))
	>>> getsent('(ROOT (S 0:2) ($. 3))', ['Foo', 'bar', 'zed', '.'])
	('(ROOT (S 0) ($. 1))', (None, '.'))
	>>> getsent('(ROOT (S 0) ($. 3))', ['Foo', 'bar', 'zed','.'])
	('(ROOT (S 0) ($. 2))', ('Foo', None, '.'))
	>>> getsent('(S|<VP>_2 (VP_3 0:1 3:3 16:16) (VAFIN 2))', '''In Japan wird
	...  offenbar die Fusion der Geldkonzerne Daiwa und Sumitomo zur
	...  gr\\xf6\\xdften Bank der Welt vorbereitet .'''.split(' '))
	('(S|<VP>_2 (VP_3 0 2 4) (VAFIN 1))', (None, 'wird', None, None, None)) """
	cdef:
		int n, m = 0, maxl
		list newsent = []
		dict leafmap = {}
		dict spans = {int(start): int(end) + 1
				for start, end in FRONTIERRE.findall(frag)}
		list leaves = list(map(int, TERMINDICESRE.findall(frag)))
	for n in leaves:
		spans[n] = n + 1
	maxl = max(spans)
	for n in sorted(spans):
		if n in leaves:
			newsent.append(sent[n])  # a terminal
		else:
			newsent.append(None)  # a frontier node
		leafmap[n] = " %d" % m
		m += 1
		if spans[n] not in spans and n != maxl:  # a gap
			newsent.append(None)
			m += 1
	frag = FRONTIERORTERMRE.sub(repl(leafmap), frag)
	return frag, tuple(newsent)


cdef dumptree(NodeArray a, Node *anodes, list asent, list labels):
	""" dump the node structs of a tree showing numeric IDs as well
	as a strings representation of the tree in bracket notation. """
	cdef int n
	for n in range(a.len):
		print('idx=%2d\tleft=%2d\tright=%2d\tprod=%2d\tlabel=%s' % (n,
				termidx(anodes[n].left) if anodes[n].left < 0
				else anodes[n].left,
				anodes[n].right, anodes[n].prod, labels[anodes[n].prod]),
				end='')
		if anodes[n].left < 0:
			if asent[termidx(anodes[n].left)]:
				print('\t%s=%s' % ('terminal', asent[termidx(anodes[n].left)]))
			else:
				print('\tfrontier non-terminal')
		else:
			print()
	print(unicodetree(anodes, labels, asent, a.root), '\n', asent, '\n')


cdef dumpCST(ULong *CST, NodeArray a, NodeArray b, Node *anodes, Node *bnodes,
		list asent, list bsent, list labels, short SLOTS, bint bitmatrix=False):
	""" dump a table of the common subtrees of two trees. """
	cdef int n, m
	cdef Node aa, bb
	dumptree(a, anodes, asent, labels)
	dumptree(b, bnodes, bsent, labels)
	print('\t'.join([''] + ["%2d" % x for x in range(b.len)
		if bnodes[x].prod != -1]))
	for m in range(b.len):
		bb = bnodes[m]
		print('\t', labels[bb.prod][:3], end='')
	for n in range(a.len):
		aa = anodes[n]
		print("\n%2d" % n, labels[aa.prod][:3], end='')
		for m in range(b.len):
			print('\t', end='')
			if bitmatrix:
				print("1" if TESTBIT(&CST[m * SLOTS], n) else ' ', end='')
			else:
				if TESTBIT(&CST[IDX(n, m, b.len, SLOTS)], a.len):
					if abitcount(&CST[IDX(n, m, b.len, SLOTS)], SLOTS) - 1:
						print(abitcount(&CST[IDX(n, m, b.len, SLOTS)],
								SLOTS) - 1, end='')
				else:
					print('-', end='')
	print("\ncommon productions:", end='')
	print(len({anodes[n].prod for n in range(a.len)} &
			{bnodes[n].prod for n in range(b.len)}))
	print("found:", end='')
	if bitmatrix:
		print("horz", sum([abitcount(&CST[n * SLOTS], SLOTS) > 0
				for n in range(b.len)]),
			"vert", sum([any([TESTBIT(&CST[n * SLOTS], m)
				for n in range(b.len)]) for m in range(a.len)]),
			"both", abitcount(CST, b.len * SLOTS))
	else:
		print(sum([any([abitcount(&CST[IDX(n, m, b.len, SLOTS)], SLOTS) > 1
			for m in range(b.len)]) for n in range(a.len)]))


def add_lcfrs_rules(tree, sent):
	""" Set attribute '.prod' on nodes of tree with the LCFRS production for
	that node. """
	for a, b in zip(tree.subtrees(),
			lcfrs_productions(tree, sent, frontiers=True)):
		a.prod = b
	return tree


def getlabelsprods(trees, labels, prods):
	""" collect label/prod attributes from 'trees' and index them. """
	pnum = len(prods)
	for tree in trees:
		for st in tree:
			if st.prod not in prods:
				labels.append(st.label)
				prods[st.prod] = pnum
				pnum += 1


def nonfrontier(sent):
	def nonfrontierfun(x):
		return isinstance(x[0], Tree) or sent[x[0]] is not None
	return nonfrontierfun


def tolist(tree, sent):
	""" Convert Tree object to list of non-terminal nodes in pre-order
	traversal; add indices to nodes reflecting their position in the list. """
	result = list(tree.subtrees(nonfrontier(sent)))
	for n in reversed(range(len(result))):
		a = result[n]
		a.idx = n
		assert a.label, ("labels should be non-empty. tree: "
				"%s\nsubtree: %s\nindex %d, label %r" % (tree, a, n, a.label))
	result[0].rootidx = 0
	return result


def pathsplit(p):
	return p.rsplit('/', 1) if '/' in p else ('.', p)


def getprodid(prods, node):
	return prods.get(node.prod, -1)


def getctrees(trees, sents, trees2=None, sents2=None):
	""" :returns: Ctrees object for disc. binary trees and sentences. """
	# make deep copies to avoid side effects.
	trees12 = trees = [tolist(add_lcfrs_rules(Tree.convert(a), b), b)
			for a, b in zip(trees, sents)]
	if trees2:
		trees2 = [tolist(add_lcfrs_rules(Tree.convert(a), b), b)
					for a, b in zip(trees2, sents2)]
		trees12 = trees + trees2
	labels = []
	prods = {}
	getlabelsprods(trees12, labels, prods)
	for tree in trees12:
		root = tree[0]
		tree.sort(key=partial(getprodid, prods))
		for n, a in enumerate(tree):
			a.idx = n
		tree[0].rootidx = root.idx
	trees = Ctrees(trees, prods)
	trees.indextrees(prods)
	if trees2:
		trees2 = Ctrees(trees2, prods)
		trees2.indextrees(prods)
	return dict(trees1=trees, sents1=sents, trees2=trees2, sents2=sents2,
			prods=prods, labels=labels)


cdef readnode(line, list labels, dict prods, Node *result,
		size_t *idx, list sent, origlabel):
	""" Parse an s-expression in a string, and store in an array of Node
	structs (pre-allocated). idx is a counters to keep track of the numer
	of Node structs used, sent collects the terminals encountered. """
	cdef:
		int n, parens = 0, start = 0, startchild2 = 0, left = -1, right = -1
		list childlabels = None
		#bytes a, label, rest, labelchild1, labelchild2
		#bytes child1 = None, child2 = None
	match = LABEL.match(line)
	if match is None:
		raise ValueError('malformed tree:\n%s' % line)
	label = match.group(1)
	rest = line[match.end():]
	child1 = child2 = labelchild2 = None
	# walk through the string and find the first two children
	for n, a in enumerate(rest):
		if a == '(':
			if parens == 0:
				start = n
			parens += 1
		elif a == ')':
			parens -= 1
			if parens == -1:
				if child1 is None:
					child1 = rest[start:n]
			elif parens == 0:
				if child1 is None:
					child1 = rest[start:n + 1]
				elif child2 is None:
					child2 = rest[start:n + 1]
					startchild2 = start
				else:  # will do on the fly binarization
					childlabels = []
					break
	# if there were more children, collect labels for a binarized constituent
	if childlabels is not None:
		for n, a in enumerate(rest[startchild2:], startchild2):
			if a == '(':
				if parens == 0:
					start = n
				parens += 1
			elif a == ')':
				parens -= 1
				if parens == 0:
					match = LABEL.match(rest, start)
					if match is None:  # introduce preterminal
						childlabels.append('/'.join((label, rest[start:n + 1])))
					else:
						childlabels.append(match.group(1))
		# not optimal, parsing this string can be avoided
		# in fact we already have: label, child2, rest
		child2 = ('(' + (origlabel or label) + '|<' + ','.join(childlabels)
				+ '> ' + rest[startchild2:])
	assert parens == -1, "unbalanced parentheses: %d\n%r" % (parens, line)
	match1 = LABEL.match(child1)
	if match1 is not None:  # non-terminal label
		labelchild1 = match1.group(1)
	elif child2 is not None:  # insert preterminal
		labelchild1 = '/'.join((label, child1))
		child1 = '(' + ' '.join((labelchild1, child1)) + ')'
	else:  # terminal following pre-terminal; store terminal
		# leading space distinguishes terminal from non-terminal
		labelchild1 = ' ' + child1
		left = termidx(len(sent))
		sent.append(child1 or None)
	if child2 is None:
		prod = (label, labelchild1) if child1 else (label, )
	else:
		match = LABEL.match(child2)
		if match is not None:
			labelchild2 = match.group(1)
		else:  # insert preterminal
			labelchild2 = '/'.join((label, child2))
			child2 = '(' + ' '.join((labelchild2, child2)) + ')'
		prod = (label, labelchild1, labelchild2)
	if prod not in prods:  # assign new ID?
		prods[prod] = len(prods)
		labels.append(label)
	n = idx[0]
	idx[0] += 1
	if match1 is not None or child2 is not None:
		left = idx[0]
		readnode(child1, labels, prods, result, idx, sent, None)
		if child2 is not None:
			right = idx[0]
			readnode(child2, labels, prods, result, idx, sent,
					childlabels and (origlabel or label))
	# store node
	result[n].prod = prods[prod]
	result[n].left = left
	result[n].right = right


def readtreebank(treebankfile, list labels, dict prods, bint sort=True,
		bint discontinuous=False, int limit=0, encoding="utf-8"):
	""" Read a treebank from a given filename. labels and prods should be a
	list and a dictionary, with the same ones used when reading multiple
	treebanks. """
	# this could use BracketCorpusReader or expect trees/sents as input,
	# but this version reads the treebank incrementally which reduces memory
	# requirements.
	cdef size_t cnt
	cdef Node *scratch
	cdef Ctrees ctrees
	if treebankfile is None:
		return None, None
	if discontinuous:
		# no incremental reading w/disc trees
		from discodop.treebank import NegraCorpusReader
		from discodop.treetransforms import canonicalize
		corpus = NegraCorpusReader(*pathsplit(treebankfile), encoding=encoding)
		trees = list(corpus.parsed_sents().values())
		sents = list(corpus.sents().values())
		if limit:
			trees = trees[:limit]
			sents = sents[:limit]
		trees = [tolist(add_lcfrs_rules(canonicalize(binarize(tree)),
				sent), sent) for tree, sent in zip(trees, sents)]
		getlabelsprods(trees, labels, prods)
		if sort:
			for tree in trees:
				root = tree[0]
				tree.sort(key=partial(getprodid, prods))
				for n, a in enumerate(tree):
					a.idx = n
				tree[0].rootidx = root.idx
		trees = Ctrees(trees, prods)
		return trees, sents
	# do incremental reading of bracket trees
	data = io.open(treebankfile, encoding=encoding).read()
	if limit:
		data = re.match(r'(?:[^\n\r]+[\n\r]+){0,%d}' % limit, data).group()
	lines = [line for line in data.splitlines() if line.strip()]
	numtrees = len(lines)
	assert numtrees, "%r appears to be empty" % treebankfile
	sents = []
	nodesperline = [line.count('(') for line in lines]
	numnodes = sum(nodesperline)
	ctrees = Ctrees()
	ctrees.alloc(numtrees, numnodes)
	ctrees.maxnodes = max(nodesperline)
	binfactor = 2  # conservative estimate to accommodate binarization
	scratch = <Node *>malloc(ctrees.maxnodes * binfactor * sizeof(Node))
	assert scratch is not NULL
	for m, line in enumerate(lines):
		cnt = 0
		sent = []
		readnode(line, labels, prods, scratch, &cnt, sent, None)
		sents.append(sent)
		ctrees.addnodes(scratch, cnt, 0)
	free(scratch)
	return ctrees, sents


def one():
	""" used to generate the value 1 for defaultdicts. """
	return 1


def main():
	treebank = [binarize(Tree(x)) for x in """\
(S (NP (DT The) (NN cat)) (VP (VBP saw) (NP (DT the) (JJ hungry) (NN dog))))
(S (NP (DT The) (NN cat)) (VP (VBP saw) (NP (DT the) (NN dog))))
(S (NP (DT The) (NN mouse)) (VP (VBP saw) (NP (DT the) (NN cat))))
(S (NP (DT The) (NN mouse)) (VP (VBP saw) (NP (DT the) (JJ yellow) (NN cat))))
(S (NP (DT The) (JJ little) (NN mouse)) (VP (VBP saw) (NP (DT the) (NN cat))))
(S (NP (DT The) (NN cat)) (VP (VBP ate) (NP (DT the) (NN dog))))
(S (NP (DT The) (NN mouse)) (VP (VBP ate) (NP (DT the) (NN cat))))\
		""".splitlines()]
	sents = [tree.leaves() for tree in treebank]
	for tree in treebank:
		for n, idx in enumerate(tree.treepositions('leaves')):
			tree[idx] = n
	params = getctrees(treebank, sents)
	fragments = fastextractfragments(params['trees1'], params['sents1'], 0, 0,
			params['labels'], discontinuous=True, approx=False)
	counts = exactcounts(params['trees1'], params['trees1'],
			list(fragments.values()), fast=True)
	assert len(fragments) == 25
	for (a, b), c in sorted(zip(fragments, counts), key=repr):
		print("%s\t%d" % (re.sub("[0-9]+", lambda x: b[int(x.group())], a), c))

if __name__ == '__main__':
	main()
